/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use apk_tools_api::{ApkChange, ApkPackage};

/**
* The state of a package
*/
#[derive(Copy, Clone, Debug, PartialEq, PartialOrd, Eq)]
pub enum PackageState {
    Available,
    Installed,
    PendingInstall,
    PendingRemoval,
    Upgradable,
    Downgradable,
    Reinstall,
}

pub struct ApkDiffPackage {
    installed_package: Option<ApkPackage>,
    staging_package: Option<ApkPackage>,
    package_state: PackageState,
}

impl ApkDiffPackage {
    pub fn from_change(change: ApkChange) -> Option<Self> {
        let installed_package = change.old_pkg();
        let staging_package = change.new_pkg();

        if let Some(ipkg) = &installed_package {
            if let Some(spkg) = &staging_package {
                if ipkg.ptr_compare(spkg) {
                    return None;
                }
            }
        }

        let mut package_state = match (&installed_package, &staging_package) {
            (Some(_), Some(_)) => {
                if installed_package > staging_package {
                    PackageState::Downgradable
                } else if installed_package < staging_package {
                    PackageState::Upgradable
                } else {
                    PackageState::Installed
                }
            }
            (Some(_), None) => PackageState::PendingRemoval,
            (None, Some(_)) => PackageState::PendingInstall,
            _ => PackageState::Available,
        };

        if change.is_reinstall() {
            package_state = PackageState::Reinstall;
        }

        Some(Self {
            installed_package,
            staging_package,
            package_state,
        })
    }

    pub fn installed_package(&self) -> Option<&ApkPackage> {
        self.installed_package.as_ref()
    }

    pub fn staging_package(&self) -> Option<&ApkPackage> {
        self.staging_package.as_ref()
    }

    pub fn package_state(&self) -> PackageState {
        self.package_state
    }
}
