/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::apk_diff_package::{ApkDiffPackage, PackageState};
use crate::details_flags::DetailsFlags;
use enumflags2::BitFlags;
use serde::Serialize;
use std::collections::{BTreeMap, HashMap, HashSet};
use std::fmt::{Debug, Formatter};
use strum::IntoEnumIterator;
#[cfg(test)]
use zvariant::Value;
use zvariant::{OwnedValue, Str};

#[derive(zvariant::Type)]
#[zvariant(signature = "dict")]
pub struct ApkPackagePartial {
    inner: HashMap<&'static str, OwnedValue>,
}

impl ApkPackagePartial {
    pub fn empty() -> Self {
        ApkPackagePartial {
            inner: HashMap::new(),
        }
    }

    pub fn from_apk_api_package(
        package: &apk_tools_api::ApkPackage,
        details_flags: BitFlags<DetailsFlags>,
    ) -> Self {
        let mut builder = builder::ApkPackagePartialBuilder::default();
        builder.name(package.name().name().to_string());

        if details_flags.contains(DetailsFlags::Description) {
            if let Some(desc) = package.description() {
                builder.description(desc.to_string());
            }
        }

        if details_flags.contains(DetailsFlags::InstalledSize) {
            builder.installed_size(package.installed_size());
        }

        if details_flags.contains(DetailsFlags::Version) {
            builder.version(package.version().to_string());
        }

        if details_flags.contains(DetailsFlags::License) {
            builder.license(package.license().to_string());
        }

        if details_flags.contains(DetailsFlags::PackageState) {
            if package.has_installed_package() {
                builder.package_state(PackageState::Installed);
            } else {
                builder.package_state(PackageState::Available);
            }
        }

        if details_flags.contains(DetailsFlags::Size) {
            builder.size(package.size());
        }

        if details_flags.contains(DetailsFlags::Url) {
            if let Some(url) = package.url() {
                builder.url(url.to_string());
            }
        }

        builder.build()
    }

    pub fn from_apk_diff_package(
        package: ApkDiffPackage,
        details_flags: BitFlags<DetailsFlags>,
    ) -> Self {
        let pkg = package
            .installed_package()
            .unwrap_or_else(|| package.staging_package().unwrap());
        let mut result = Self::from_apk_api_package(pkg, details_flags);

        if details_flags.contains(DetailsFlags::PackageState) {
            // TODO
            result.inner.insert(
                DetailsFlags::PackageState.as_ref(),
                (package.package_state() as u32).into(),
            );
        }

        if details_flags.contains(DetailsFlags::StagingVersion) {
            match package.package_state() {
                PackageState::Reinstall | PackageState::Upgradable | PackageState::Downgradable => {
                    if let Some(staging_package) = package.staging_package() {
                        result.inner.insert(
                            DetailsFlags::StagingVersion.as_ref(),
                            Str::from(staging_package.version()).into(),
                        );
                    }
                }
                _ => {}
            }
        }

        result
    }

    // TODO: This really shouldn't be in apk_package_partial
    pub fn from_error(pkgname: &str, error: &str) -> Self {
        let mut h = HashMap::new();
        h.insert("name", Str::from(pkgname).into());
        h.insert("error", Str::from(error).into());
        Self { inner: h }
    }
}

impl PartialEq for ApkPackagePartial {
    fn eq(&self, other: &Self) -> bool {
        let detail_strs = DetailsFlags::iter().map(|f| f.to_string());
        let name_property = ["name".to_string()].into_iter();
        let unstable_properties = HashSet::from(["size", "installed_size"]);

        for property in name_property
            .chain(detail_strs)
            .filter(|x| !unstable_properties.contains(x.as_str()))
        {
            if self.inner.get(property.as_str()) != other.inner.get(property.as_str()) {
                return false;
            }
        }

        true
    }
}

impl Serialize for ApkPackagePartial {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        self.inner.serialize(serializer)
    }
}

impl Debug for ApkPackagePartial {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let ordered_inner: BTreeMap<&&'static str, &OwnedValue> = self.inner.iter().collect();
        f.debug_struct("ApkPackagePartial")
            .field("inner", &ordered_inner)
            .finish()
    }
}

#[cfg(test)]
impl ApkPackagePartial {
    pub fn builder() -> builder::ApkPackagePartialBuilder {
        builder::ApkPackagePartialBuilder::default()
    }

    pub fn is_err(&self) -> bool {
        self.inner.contains_key("error")
    }

    pub fn err(&self) -> String {
        let error = Value::downcast_ref::<Str>(self.inner.get("error").unwrap()).unwrap();
        let binding = error.as_str();
        String::from(binding)
    }
}

mod builder {
    use super::ApkPackagePartial;
    use crate::apk_diff_package::PackageState;
    use std::collections::HashMap;
    use zvariant::Str;

    #[derive(Debug, Clone, Default)]
    pub struct ApkPackagePartialBuilder {
        name: Option<String>,
        description: Option<String>,
        installed_size: Option<u64>,
        version: Option<String>,
        license: Option<String>,
        package_state: Option<PackageState>,
        size: Option<u64>,
        staging_version: Option<String>,
        url: Option<String>,
    }

    impl ApkPackagePartialBuilder {
        pub fn name(&mut self, v: String) -> &mut Self {
            self.name = Some(v);
            self
        }

        pub fn description(&mut self, v: String) -> &mut Self {
            self.description = Some(v);
            self
        }

        pub fn installed_size(&mut self, v: u64) -> &mut Self {
            self.installed_size = Some(v);
            self
        }

        pub fn version(&mut self, v: String) -> &mut Self {
            self.version = Some(v);
            self
        }

        pub fn license(&mut self, v: String) -> &mut Self {
            self.license = Some(v);
            self
        }

        pub fn package_state(&mut self, v: PackageState) -> &mut Self {
            self.package_state = Some(v);
            self
        }

        pub fn size(&mut self, v: u64) -> &mut Self {
            self.size = Some(v);
            self
        }

        #[allow(dead_code)]
        pub fn staging_version(&mut self, v: String) -> &mut Self {
            self.staging_version = Some(v);
            self
        }

        pub fn url(&mut self, v: String) -> &mut Self {
            self.url = Some(v);
            self
        }

        pub fn build(self) -> ApkPackagePartial {
            let mut h = HashMap::new();
            if let Some(v) = self.name {
                h.insert("name", Str::from(v).into());
            }
            if let Some(v) = self.description {
                h.insert("description", Str::from(v).into());
            }
            if let Some(v) = self.installed_size {
                h.insert("installed_size", v.into());
            }
            if let Some(v) = self.version {
                h.insert("version", Str::from(v).into());
            }
            if let Some(v) = self.license {
                h.insert("license", Str::from(v).into());
            }
            if let Some(v) = self.package_state {
                h.insert("package_state", (v as u32).into());
            }
            if let Some(v) = self.size {
                h.insert("size", v.into());
            }
            if let Some(v) = self.staging_version {
                h.insert("staging_version", Str::from(v).into());
            }
            if let Some(v) = self.url {
                h.insert("url", Str::from(v).into());
            }
            ApkPackagePartial { inner: h }
        }
    }
}

#[cfg(test)]
mod test {
    use super::ApkPackagePartial;
    use zvariant::Type;

    #[test]
    fn signature() {
        assert_eq!(ApkPackagePartial::SIGNATURE, "a{sv}");
    }

    #[test]
    fn size_isnt_used_in_eq() {
        let mut builder_a = ApkPackagePartial::builder();
        let mut builder_b = ApkPackagePartial::builder();

        builder_a.size(5);
        builder_b.size(6);

        assert_eq!(builder_a.build(), builder_b.build());
    }
}
