/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use anyhow::{Result, anyhow};
use log::{debug, info, trace};
use serde::{
    Serialize,
    ser::{SerializeStruct, Serializer},
};
use std::fmt::{self, Write};
use tokio::io::AsyncReadExt;
use tokio::{fs::File, io::AsyncWriteExt};

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct ApkRepository {
    pub enabled: bool,
    pub description: Option<String>,
    pub url: String,
}

impl ApkRepository {
    pub fn new(repo_line: &str, description: Option<String>) -> Self {
        let enabled = !repo_line.starts_with('#');
        let url = repo_line.replace('#', "").trim().to_string();
        ApkRepository {
            enabled,
            description,
            url,
        }
    }

    /**
     * Gets all available repositories in `$root/etc/apk/repositories`.
     *
     * # Errors
     * * Returns an `Error::GetRepositories` if opening the file or reading from it fails.
     */
    pub async fn get_repositories(root: Option<&str>) -> Result<Vec<Self>> {
        info!("Getting repositories");
        let repo_file_path = if let Some(root) = root {
            format!("{root}/etc/apk/repositories")
        } else {
            "/etc/apk/repositories".to_string()
        };
        let mut file = File::open(&repo_file_path).await.map_err(|e| {
            anyhow!("Failed to get repositories from file {repo_file_path} due to error {e}")
        })?;
        let mut repo_string = String::new();
        file.read_to_string(&mut repo_string).await?;

        let mut result = Vec::new();

        for line in repo_string.lines() {
            let trimmed = line.trim();
            if !trimmed.is_empty() {
                trace!("Creating repository from line {}", line);
                result.push(ApkRepository::new(trimmed, Some(repo_file_path.clone())));
            }
        }

        debug!(
            "Done listing repositories. Found following repositories: {:?}",
            result
        );
        Ok(result)
    }

    /**
     * Sets the repositories `repos` in the repo file `$root/etc/apk/repositories`.
     *
     * # Errors
     * * Returns an `Error::SetRepositories` if the file can't be opened or we can't write to it.
     */
    pub async fn set_repositories(repos: &[ApkRepository], root: Option<&str>) -> Result<()> {
        info!("Setting repositories to {:?}", repos);

        let mut line = String::new();
        let repo_file_path = if let Some(root) = root {
            format!("{root}/etc/apk/repositories")
        } else {
            "/etc/apk/repositories".to_string()
        };
        let mut file = File::create(&repo_file_path).await.map_err(|e| {
            anyhow!("Failed to set repositories from file {repo_file_path} due to error {e}",)
        })?;
        for repo in repos {
            trace!("Writing repository {} to file {}", repo, repo_file_path);
            if !repo.enabled {
                line.push('#');
            }
            writeln!(line, "{}", repo.url)?;
        }
        file.write_all(line.as_bytes()).await?;
        file.flush().await?;

        debug!("Done setting repositories to {:?}", repos);
        Ok(())
    }
}

impl fmt::Display for ApkRepository {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let enabled_str = if self.enabled { "enabled" } else { "disabled" };
        write!(
            f,
            "{} repository {} originating from file {:?}",
            enabled_str, self.url, self.description
        )
    }
}

impl Serialize for ApkRepository {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut state = serializer.serialize_struct("ApkRepository", 3)?;
        state.serialize_field("enabled", &self.enabled)?;
        state.serialize_field(
            "description",
            &self.description.as_ref().unwrap_or(&"".to_string()),
        )?;
        state.serialize_field("url", &self.url)?;
        state.end()
    }
}

#[cfg(test)]
mod test {
    use super::ApkRepository;
    use serde_test::{Token, assert_ser_tokens};

    #[test]
    fn test_apk_repository_serialize() {
        let repo = ApkRepository {
            enabled: true,
            description: Some("/etc/apk/repositories".to_string()),
            url: "https://example.org".to_string(),
        };
        assert_ser_tokens(
            &repo,
            &[
                Token::Struct {
                    name: "ApkRepository",
                    len: 3,
                },
                Token::Str("enabled"),
                Token::Bool(true),
                Token::Str("description"),
                Token::Str("/etc/apk/repositories"),
                Token::Str("url"),
                Token::Str("https://example.org"),
                Token::StructEnd,
            ],
        );
    }

    #[test]
    fn test_apk_repository_serialize_no_description() {
        let repo = ApkRepository {
            enabled: true,
            description: None,
            url: "https://example.org".to_string(),
        };
        assert_ser_tokens(
            &repo,
            &[
                Token::Struct {
                    name: "ApkRepository",
                    len: 3,
                },
                Token::Str("enabled"),
                Token::Bool(true),
                Token::Str("description"),
                Token::Str(""),
                Token::Str("url"),
                Token::Str("https://example.org"),
                Token::StructEnd,
            ],
        );
    }
}
