/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::components::PolkitAuthenticator;
use enumflags2::BitFlags;
use std::collections::HashMap;
use zbus::Connection;
use zbus::fdo::Error;
use zbus::message::Header;
use zbus_polkit::policykit1::{
    AuthorityProxy, AuthorizationResult, CheckAuthorizationFlags, Subject,
};

#[derive(shaku::Component)]
#[shaku(interface = PolkitAuthenticator)]
pub struct DbusPolkitAuthenticator;

#[async_trait::async_trait]
impl PolkitAuthenticator for DbusPolkitAuthenticator {
    async fn check_polkit_auth(
        &self,
        connection: &Connection,
        header: &Header<'_>,
    ) -> Result<(), Error> {
        let proxy = AuthorityProxy::new(connection).await?;
        let subject = Self::get_message_header_subject(header)?;
        let check_auth_details = self.get_auth_details();
        let fully_qualified_action_id = Self::get_fully_qualified_action_id(header)?;
        let res = Self::proxy_check_auth(
            proxy,
            &subject,
            &fully_qualified_action_id,
            &check_auth_details,
        )
        .await?;

        if res.is_authorized {
            Ok(())
        } else {
            Err(Error::AuthFailed(format!(
                "Access for operation {fully_qualified_action_id} was denied",
            )))
        }
    }
}

impl DbusPolkitAuthenticator {
    fn get_message_header_subject(header: &Header<'_>) -> Result<Subject, Error> {
        Ok(match Subject::new_for_message_header(header) {
            Ok(o) => o,
            Err(e) => {
                return Err(Error::AuthFailed(format!(
                    "Failed to determine subject: {e}",
                )));
            }
        })
    }

    fn get_auth_details(&self) -> HashMap<&str, &str> {
        let mut check_auth_details = HashMap::new();
        check_auth_details.insert("polkit.gettext_domain", "apk-polkit-rs");
        check_auth_details
    }

    fn get_fully_qualified_action_id(header: &Header<'_>) -> Result<String, Error> {
        match header.member().map(|m| m.as_str().to_string()) {
            Some(id) => Ok(format!("dev.Cogitri.apkPolkit2.{id}")),
            None => Err(Error::AuthFailed(
                "Failed to determine action ID".to_string(),
            )),
        }
    }

    async fn proxy_check_auth(
        proxy: AuthorityProxy<'_>,
        subject: &Subject,
        fully_qualified_action_id: &str,
        check_auth_details: &HashMap<&str, &str>,
    ) -> Result<AuthorizationResult, Error> {
        proxy
            .check_authorization(
                subject,
                fully_qualified_action_id,
                check_auth_details,
                BitFlags::from(CheckAuthorizationFlags::AllowUserInteraction),
                "",
            )
            .await
            .map_err(|e| Error::AuthFailed(format!("Failed to query for authentication: {e}",)))
    }
}
