/*
   Copyright: 2020-2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/
use crate::apk_diff_package::ApkDiffPackage;
use crate::apk_repositories::ApkRepository;
#[cfg(test)]
use crate::components::TestPolkitAuthenticator;
use crate::{
    apk_package_partial::ApkPackagePartial,
    change_repository_actions::ChangeRepositoryActions,
    components::{ComponentBroker, PolkitAuthenticator},
    details_flags::DetailsFlags,
    error,
};
use anyhow::anyhow;
use apk_tools_api::{ApkCtx, ApkCtxBuilder, ApkDatabase, ApkDependency, ApkDependencyArray};
use enumflags2::BitFlags;
use log::{Level, debug, info};
use serde::Serialize;
use shaku::HasComponent;
use std::fs;
use std::path::Path;
use std::process::Command;
use zbus::{self, Connection, message::Header};

pub struct DBusServer {
    #[cfg(test)]
    additional_repo_uri: Option<String>,
    root: Option<String>,
    component_broker: ComponentBroker,
}

#[derive(Debug, Serialize, zvariant::Type, PartialEq, Eq)]
#[zvariant(signature = "(bss)")]
pub struct ApkRepositoryDBus(ApkRepository);

#[cfg(test)]
impl DBusServer {
    pub fn new_explicit(additional_repo_uri: Option<String>, root: Option<String>) -> Self {
        let component_broker = ComponentBroker::builder()
            .with_component_override(Box::new(TestPolkitAuthenticator::new()))
            .build();

        Self {
            additional_repo_uri,
            root,
            component_broker,
        }
    }
}

impl DBusServer {
    pub fn new() -> Self {
        Self {
            #[cfg(test)]
            additional_repo_uri: Default::default(),
            root: Default::default(),
            component_broker: ComponentBroker::builder().build(),
        }
    }

    fn get_apk_ctx(&self, read_only: bool) -> ApkCtxBuilder {
        let mut builder = ApkCtx::builder();
        builder.set_read_only(read_only);

        if let Some(root) = &self.root {
            builder.set_root(root.clone());
        }

        #[cfg(test)]
        {
            builder.set_allow_untrusted(true);
            builder.set_usermode(true);
            if let Some(uri) = &self.additional_repo_uri {
                builder.add_additional_repository(uri);
            }
        }

        builder
    }

    async fn check_polkit_auth(
        &self,
        header: &Header<'_>,
        connection: &Connection,
    ) -> Result<(), zbus::fdo::Error> {
        let authenticator: &dyn PolkitAuthenticator = self.component_broker.resolve_ref();
        authenticator.check_polkit_auth(connection, header).await
    }

    fn get_package_details(
        db: &mut ApkDatabase,
        pkgname: &str,
    ) -> anyhow::Result<apk_tools_api::ApkPackage> {
        let dependency = ApkDependency::from_pkgname(db, pkgname)?;
        if let Some(dependency) = dependency {
            if let Some(providers) = dependency.name().providers() {
                if let Some(first_provider) = providers.iter().next() {
                    if let Some(pkg) = first_provider.pkg() {
                        if log::log_enabled!(Level::Debug) {
                            let name = pkg.name();
                            debug!("Resolved pkg {pkgname} to {name}");
                        }
                        return Ok(pkg);
                    }
                }
            }
        }

        Err(anyhow!(
            "Couldn't retrieve package details for package {pkgname}"
        ))
    }

    fn resolve_symlink(&self, path: &str) -> anyhow::Result<String> {
        let file_path = if let Some(ref root) = self.root {
            let path = Path::new(path);
            let normalized = if path.has_root() {
                Path::new(path).strip_prefix("/")?.to_path_buf()
            } else {
                Path::new(root).join(path).to_path_buf()
            };
            let absolute_path = Path::new(root).join(normalized);
            let canonlized_path = fs::canonicalize(absolute_path)?;
            canonlized_path.strip_prefix(root)?.to_path_buf()
        } else {
            fs::canonicalize(path)?
        };

        let file_path_str = file_path
            .to_str()
            .ok_or(anyhow!("Couldn't convert path to string!"))?;

        debug!("Resolved {path} to {file_path_str}");

        Ok(file_path_str.to_string())
    }

    fn set_update_flag_for_main_package(
        db: &mut ApkDatabase,
        pkgname: &str,
        dependency: ApkDependency,
    ) -> anyhow::Result<()> {
        let package = dependency
            .name()
            .get_installed_package()
            .ok_or(anyhow!("Couldn't find installed package for {pkgname}"))?;
        let origin = package
            .origin()
            .ok_or(anyhow!("Installed package {pkgname} doesn't have origin"))?;

        if origin == pkgname {
            return Ok(());
        }

        if let Some(mut origin_name) = db.query_name(&origin) {
            let is_installed = origin_name.get_installed_package().is_some();
            if is_installed {
                let dep = ApkDependency::from_pkgname(db, &origin)?.ok_or(anyhow!(
                    "Couldn't upgrade unknown transitive dependency {origin} for {pkgname}"
                ))?;
                dep.name().set_update_flags();
            }
        }

        Ok(())
    }
}

#[zbus::interface(name = "dev.Cogitri.apkPolkit2")]
impl DBusServer {
    async fn add_packages(
        &mut self,
        #[zbus(header)] header: Header<'_>,
        #[zbus(connection)] connection: &Connection,
        packages: Vec<&str>,
    ) -> error::Result<()> {
        info!("Adding packages {:?}", packages);
        self.check_polkit_auth(&header, connection).await?;

        let mut db = ApkDatabase::new();
        db.open(self.get_apk_ctx(false).build())?;

        let mut world = db.get_world();
        for pkgname in packages {
            if let Some(dependency) = ApkDependency::from_pkgname(&mut db, pkgname)? {
                world.push(dependency);
            } else {
                return Err(anyhow!("Couldn't add unknown package {pkgname}"))?;
            }
        }

        db.commit(&mut world)?;

        Ok(())
    }

    async fn add_repository(
        &mut self,
        #[zbus(header)] header: Header<'_>,
        #[zbus(connection)] connection: &Connection,
        url: &str,
    ) -> error::Result<()> {
        info!("Add repository {url}");
        self.check_polkit_auth(&header, connection).await?;
        let mut repos = ApkRepository::get_repositories(self.root.as_deref()).await?;
        if !repos.iter().any(|r| r.url == url) {
            repos.push(ApkRepository::new(url, None));
            ApkRepository::set_repositories(&repos, self.root.as_deref()).await?;
        }
        Ok(())
    }

    async fn delete_packages(
        &mut self,
        #[zbus(header)] header: Header<'_>,
        #[zbus(connection)] connection: &Connection,
        packages: Vec<&str>,
    ) -> error::Result<()> {
        info!("Deleting packages {:?}", packages);
        self.check_polkit_auth(&header, connection).await?;

        let mut db = ApkDatabase::new();
        db.open(self.get_apk_ctx(false).build())?;

        let mut world = db.get_world();
        for pkgname in &packages {
            if let Some(dependency) = ApkDependency::from_pkgname(&mut db, pkgname)? {
                world.delete(&dependency);
            } else {
                return Err(anyhow!("Couldn't delete unknown package {pkgname}"))?;
            }
        }

        let mut changeset = db.solve_delete_changeset(&mut world)?;
        changeset.mark_new_packages();

        for pkgname in packages {
            if let Some(reverse_dependencies) = db.get_reverse_dependencies(pkgname) {
                let mut deps: Vec<String> = reverse_dependencies.into_iter().collect();
                deps.sort();
                return Err(anyhow!(
                    "Package {pkgname} is still required by the following packages: {}",
                    deps.join(" ")
                ))?;
            }
        }

        db.commit_changeset(&mut changeset, &mut world)?;

        Ok(())
    }

    async fn get_packages_details(
        &self,
        requested_packages: Vec<&str>,
        requested_properties: BitFlags<DetailsFlags>,
    ) -> error::Result<Vec<ApkPackagePartial>> {
        info!("Getting package details for {:?}", requested_packages);
        let mut db = ApkDatabase::new();
        db.open(self.get_apk_ctx(false).build())?;
        let mut result = Vec::new();

        for pkgname in requested_packages {
            let val = match Self::get_package_details(&mut db, pkgname) {
                Ok(pkg) => ApkPackagePartial::from_apk_api_package(&pkg, requested_properties),
                Err(e) => ApkPackagePartial::from_error(pkgname, &e.to_string()),
            };
            result.push(val);
        }

        Ok(result)
    }

    async fn list_repositories(&self) -> error::Result<Vec<ApkRepositoryDBus>> {
        info!("Listing repositories");
        Ok(ApkRepository::get_repositories(self.root.as_deref())
            .await?
            .drain(..)
            .map(ApkRepositoryDBus)
            .collect())
    }

    async fn list_upgradable_packages(
        &self,
        requested_properties: BitFlags<DetailsFlags>,
    ) -> error::Result<Vec<ApkPackagePartial>> {
        info!("Listing upgradable packages");
        let mut db = ApkDatabase::new();
        db.open(self.get_apk_ctx(true).build())?;

        db.check_world()?;
        let mut world = db.get_world();
        for mut dependency in world.iter() {
            if dependency.has_checksum_depmask() {
                dependency.reset_for_upgrade();
            }
        }
        let changeset = db.solve_upgrade_changeset(&mut world)?;
        let changes = changeset.changes().unwrap();

        let result = changes
            .iter()
            .filter_map(ApkDiffPackage::from_change)
            .map(|p| ApkPackagePartial::from_apk_diff_package(p, requested_properties))
            .collect();

        Ok(result)
    }

    async fn remove_repository(
        &mut self,
        #[zbus(header)] header: Header<'_>,
        #[zbus(connection)] connection: &Connection,
        url: &str,
    ) -> error::Result<()> {
        info!("Removing repository {url}");
        self.check_polkit_auth(&header, connection).await?;
        let mut repos = ApkRepository::get_repositories(self.root.as_deref()).await?;
        repos.retain(|r| r.url != url);
        ApkRepository::set_repositories(&repos, self.root.as_deref()).await?;
        Ok(())
    }

    async fn change_repository(
        &mut self,
        #[zbus(header)] header: Header<'_>,
        #[zbus(connection)] connection: &Connection,
        url: &str,
        action: i32,
    ) -> error::Result<()> {
        info!("Changing repository {url} (action = {action})");
        self.check_polkit_auth(&header, connection).await?;

        let action = ChangeRepositoryActions::try_from(action)?;
        let mut repos = ApkRepository::get_repositories(self.root.as_deref()).await?;
        if let Some(r) = repos.iter_mut().find(|r| r.url == url) {
            r.enabled = action == ChangeRepositoryActions::Enable;
        }
        ApkRepository::set_repositories(&repos, self.root.as_deref()).await?;
        Ok(())
    }

    async fn search_files_owners(
        &self,
        paths: Vec<&str>,
        requested_properties: BitFlags<DetailsFlags>,
    ) -> error::Result<Vec<ApkPackagePartial>> {
        info!("Searching for file owner(s) of {:?}", paths);
        let mut db = ApkDatabase::new();
        db.open(self.get_apk_ctx(true).build())?;
        let mut result = Vec::new();

        for path in paths {
            let owner_pkg = self
                .resolve_symlink(path)
                .ok()
                .and_then(|path| db.get_file_owner(&path));
            result.push(owner_pkg.map_or_else(ApkPackagePartial::empty, |p| {
                ApkPackagePartial::from_apk_api_package(&p, requested_properties)
            }));
        }

        Ok(result)
    }

    async fn update_repositories(
        &mut self,
        #[zbus(header)] header: Header<'_>,
        #[zbus(connection)] connection: &Connection,
    ) -> error::Result<()> {
        info!("Updating repositories");
        self.check_polkit_auth(&header, connection).await?;

        let mut db = ApkDatabase::new();
        db.open(self.get_apk_ctx(false).build())?;

        db.update_repositories()?;

        Command::new("alpine-appstream-downloader").output()?;

        Ok(())
    }

    async fn upgrade_packages(
        &mut self,
        #[zbus(header)] header: Header<'_>,
        #[zbus(connection)] connection: &Connection,
        pkgnames: Vec<&str>,
    ) -> error::Result<()> {
        info!("Upgrading packages {:?}", pkgnames);
        self.check_polkit_auth(&header, connection).await?;

        let mut db = ApkDatabase::new();
        db.open(self.get_apk_ctx(false).build())?;

        for pkgname in pkgnames {
            let dependency = ApkDependency::from_pkgname(&mut db, pkgname)?
                .ok_or(anyhow!("Couldn't upgrade unknown package {pkgname}"))?;

            dependency.name().set_update_flags();

            Self::set_update_flag_for_main_package(&mut db, pkgname, dependency)?;
        }

        let mut world = db.get_world();
        db.commit(&mut world)?;

        Ok(())
    }

    async fn cache_packages(
        &mut self,
        #[zbus(header)] header: Header<'_>,
        #[zbus(connection)] connection: &Connection,
        pkgnames: Vec<&str>,
        repositories_file: &str,
        cache_world: bool,
    ) -> error::Result<()> {
        info!(
            "Caching packages {:?} (repository_file = {repositories_file}, cache_world = {cache_world})",
            pkgnames
        );
        self.check_polkit_auth(&header, connection).await?;

        let mut db = ApkDatabase::new();
        let mut builder = self.get_apk_ctx(false);
        builder.set_repositories_file(repositories_file.to_string());
        db.open(builder.build())?;
        db.update_repositories()?;
        db.check_world()?;

        let mut dependencies = if cache_world {
            db.get_world()
        } else {
            ApkDependencyArray::new()
        };

        for pkgname in pkgnames {
            if let Some(dependency) = ApkDependency::from_pkgname(&mut db, pkgname)? {
                dependencies.push(dependency);
            } else {
                return Err(anyhow!("Couldn't cache unknown package {pkgname}"))?;
            }
        }

        for mut dependency in dependencies.iter() {
            if dependency.has_checksum_depmask() {
                dependency.reset_for_upgrade();
            }
        }
        let changeset = db.solve_upgrade_changeset(&mut dependencies)?;
        let changes = changeset.changes().unwrap();

        for change in changes.iter() {
            if let Some(mut pkg) = change.new_pkg() {
                if let Some(mut repo) = db.select_repo(&mut pkg) {
                    db.cache_package(&mut repo, &mut pkg)?;
                }
            }
        }

        Ok(())
    }
}

#[cfg(test)]
mod test {
    use super::DBusServer;
    use crate::apk_diff_package::PackageState;
    use crate::apk_repositories::ApkRepository;
    use crate::{
        apk_package_partial::ApkPackagePartial, dbus_server::ChangeRepositoryActions,
        details_flags::DetailsFlags,
    };
    use enumflags2::BitFlag;
    use std::fs::read_dir;
    use std::io::Write;
    use std::path::{Path, PathBuf};
    use std::process::Command;
    use std::sync::Once;
    use tokio::{
        fs::{File, create_dir_all},
        io::AsyncWriteExt,
    };
    use zbus::*;

    static INIT_LOGGER: Once = Once::new();

    fn start_fs_test() -> (String, String, tempfile::TempDir) {
        INIT_LOGGER.call_once(|| {
            fern::Dispatch::new()
                .format(move |out, message, record| {
                    out.finish(format_args!("[{}] {}", record.level(), message,))
                })
                .chain(std::io::stdout())
                .apply()
                .unwrap();
        });

        let temp_dir = tempfile::tempdir().unwrap();
        let mut repo_build_script_path = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
        repo_build_script_path.pop();
        let repo_build_script_path = repo_build_script_path
            .join("tests")
            .join("build_repo.sh")
            .to_str()
            .unwrap()
            .to_string();

        let apk_root_dir = temp_dir.as_ref().join("apk_root");
        let abuild_build_dir = temp_dir.as_ref().join("abuild_root");

        let abuild_cmd = Command::new(repo_build_script_path)
            .args(&[&apk_root_dir, &abuild_build_dir])
            .env("APK", "apk")
            .output()
            .expect("Failed to execute abuild!");

        if !abuild_cmd.status.success() {
            println!("Running abuild failed!");
            std::io::stdout().write(&abuild_cmd.stdout).unwrap();
            std::io::stdout().write(&abuild_cmd.stderr).unwrap();
            panic!("Running abuild failed!");
        }

        (
            apk_root_dir.to_str().unwrap().to_string(),
            abuild_build_dir
                .join("abuilds")
                .to_str()
                .unwrap()
                .to_string(),
            temp_dir,
        )
    }

    fn build_additional_package(pkgname: &str, repo_dir: &str, root_dir: &str) {
        let repo_path = Path::new(repo_dir).join(pkgname);

        //    APK="$APK --allow-untrusted --root $1" SUDO_APK="abuild-apk --root $1" REPODEST="$2"
        let abuild_cmd = Command::new("abuild")
            .args(&[
                "-F",
                "clean",
                "unpack",
                "prepare",
                "build",
                "rootpkg",
                "update_abuildrepo_index",
            ])
            .env("APK", format!("apk --allow-untrusted --root {}", root_dir))
            .env("SUDO_APK", format!("abuild-apk --root {}", root_dir))
            .env("REPODEST", &Path::new(repo_dir).parent().unwrap())
            .env(
                "ABUILD_USERDIR",
                &Path::new(repo_dir).parent().unwrap().join("abuildUserDir"),
            )
            .current_dir(repo_path)
            .output()
            .expect("Failed to execute abuild!");

        if !abuild_cmd.status.success() {
            println!("Running abuild failed!");
            std::io::stdout().write_all(&abuild_cmd.stdout).unwrap();
            std::io::stdout().write_all(&abuild_cmd.stderr).unwrap();
            panic!("Running abuild failed!");
        }
    }

    fn get_message() -> Message {
        let mut builder = Message::method_call("/dev/Cogitri/apkPolkit2", "AddPackages").unwrap();
        builder = builder.sender(":org.donot").unwrap();
        builder.build(&()).unwrap()
    }

    #[tokio::test]
    async fn test_add() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        server
            .add_packages(
                get_message().header(),
                &connection,
                vec!["test-a", "test-b"],
            )
            .await
            .unwrap();
        let output = Command::new(format!("{}/usr/bin/test-a", root)).output().expect("Running test-a executable failed; apparently installing the test-a package didn't work!");
        assert!(output.status.success());
        assert_eq!(output.stdout, b"hello from test-a-1.0\n");
        let output = Command::new(format!("{}/usr/bin/test-b", root)).output().expect("Running test-b executable failed; apparently installing the test-b package didn't work!");
        assert!(output.status.success());
        assert_eq!(output.stdout, b"hello from test-b-1.0\n");

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_add_unknown_package() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        assert!(
            server
                .add_packages(get_message().header(), &connection, vec!["unknown"])
                .await
                .is_err()
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_delete() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        server
            .add_packages(
                get_message().header(),
                &connection,
                vec!["test-a", "test-b"],
            )
            .await
            .unwrap();
        let a_path = Path::new(&root).join("usr").join("bin").join("test-a");
        assert!(a_path.exists());
        let b_path = Path::new(&root).join("usr").join("bin").join("test-b");
        assert!(b_path.exists());
        server
            .delete_packages(
                get_message().header(),
                &connection,
                vec!["test-a", "test-b"],
            )
            .await
            .unwrap();
        assert!(!a_path.exists());
        assert!(!b_path.exists());

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_delete_unknown_package() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        server
            .delete_packages(get_message().header(), &connection, vec!["test-a"])
            .await
            .unwrap();
        let exe_path = Path::new(&root).join("usr").join("bin").join("test-a");
        assert!(!exe_path.exists());

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_repo_add() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        let repo_file_path = Path::new(&root)
            .join("etc")
            .join("apk")
            .join("repositories");
        create_dir_all(repo_file_path.parent().unwrap())
            .await
            .unwrap();
        let mut repo_file = File::create(&repo_file_path).await.unwrap();
        repo_file.write(b"https://alpine.global.ssl.fastly.net/alpine/edge/main\n#https://alpine.global.ssl.fastly.net/alpine/edge/testing\nhttps://alpine.global.ssl.fastly.net/alpine/edge/community\n").await.unwrap();
        server
            .add_repository(
                get_message().header(),
                &connection,
                "https://alpine.global.ssl.fastly.net/alpine/edge/testing",
            )
            .await
            .unwrap();
        let repos = server.list_repositories().await.unwrap();
        let repo_file_path_s = repo_file_path.to_str().map(|s| s.to_string());
        assert_eq!(
            repos[0].0,
            ApkRepository::new(
                "https://alpine.global.ssl.fastly.net/alpine/edge/main",
                repo_file_path_s.clone(),
            )
        );
        assert_eq!(
            repos[1].0,
            ApkRepository::new(
                "# https://alpine.global.ssl.fastly.net/alpine/edge/testing",
                repo_file_path_s.clone(),
            ),
        );
        assert_eq!(
            repos[2].0,
            ApkRepository::new(
                "https://alpine.global.ssl.fastly.net/alpine/edge/community",
                repo_file_path_s.clone(),
            ),
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_repo_remove() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        let repo_file_path = Path::new(&root)
            .join("etc")
            .join("apk")
            .join("repositories");
        create_dir_all(repo_file_path.parent().unwrap())
            .await
            .unwrap();
        let mut repo_file = File::create(&repo_file_path).await.unwrap();
        repo_file.write(b"https://alpine.global.ssl.fastly.net/alpine/edge/main\n#https://alpine.global.ssl.fastly.net/alpine/edge/testing\nhttps://alpine.global.ssl.fastly.net/alpine/edge/community\n").await.unwrap();
        server
            .remove_repository(
                get_message().header(),
                &connection,
                "https://alpine.global.ssl.fastly.net/alpine/edge/community",
            )
            .await
            .unwrap();
        let repos = server.list_repositories().await.unwrap();
        let repo_file_path_s = repo_file_path.to_str().map(|s| s.to_string());
        assert_eq!(
            repos[0].0,
            ApkRepository::new(
                "https://alpine.global.ssl.fastly.net/alpine/edge/main",
                repo_file_path_s.clone(),
            )
        );
        assert_eq!(
            repos[1].0,
            ApkRepository::new(
                "# https://alpine.global.ssl.fastly.net/alpine/edge/testing",
                repo_file_path_s.clone(),
            ),
        );
        assert_eq!(repos.len(), 2);

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_update_repo() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        server
            .update_repositories(get_message().header(), &connection)
            .await
            .unwrap();

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_search_file_owner() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        server
            .add_packages(get_message().header(), &connection, vec!["test-e"])
            .await
            .unwrap();
        let mut builder = ApkPackagePartial::builder();
        builder
            .name("test-e".to_string())
            .version("1.0-r0".to_string())
            .description("Package E for apk-tools testsuite".to_string())
            .license("GPL".to_string())
            .url("http://alpinelinux.org".to_string())
            .installed_size(4096)
            .size(1238)
            .package_state(PackageState::Installed);

        assert_eq!(
            server
                .search_files_owners(vec!["/usr/bin/test-e"], DetailsFlags::all())
                .await
                .unwrap()[0],
            builder.build(),
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_list_upgradable() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo.clone()), Some(root.clone()));

        server
            .add_packages(
                get_message().header(),
                &connection,
                vec!["test-a", "test-b", "test-e"],
            )
            .await
            .unwrap();

        build_additional_package("test-a-new", &repo, &root);

        let packages = server
            .list_upgradable_packages(DetailsFlags::all())
            .await
            .unwrap();
        assert_eq!(packages.len(), 1);

        let mut builder = ApkPackagePartial::builder();
        builder
            .name("test-a".to_string())
            .version("1.0-r0".to_string())
            .staging_version("1.1-r0".to_string())
            .description("Package A for apk-tools testsuite".to_string())
            .license("GPL".to_string())
            .url("http://alpinelinux.org".to_string())
            .installed_size(4096)
            .size(1254)
            .package_state(PackageState::Upgradable);
        assert_eq!(packages[0], builder.build());

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_upgrade_package() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo.clone()), Some(root.clone()));

        server
            .add_packages(
                get_message().header(),
                &connection,
                vec!["test-a", "test-e"],
            )
            .await
            .unwrap();
        let output = Command::new(format!("{}/usr/bin/test-a", root)).output().expect("Running test-a executable failed; apparently installing the test-a package didn't work!");
        assert!(output.status.success());
        assert_eq!(output.stdout, b"hello from test-a-1.0\n");

        build_additional_package("test-a-new", &repo, &root);

        // No change should happen here
        server
            .upgrade_packages(get_message().header(), &connection, vec!["test-e"])
            .await
            .unwrap();
        let output = Command::new(format!("{}/usr/bin/test-a", root)).output().expect("Running test-a executable failed; apparently installing the test-a package didn't work!");
        assert!(output.status.success());
        assert_eq!(output.stdout, b"hello from test-a-1.0\n");

        // Actually upgrade, now it should be 1.1 not 1.0
        server
            .upgrade_packages(get_message().header(), &connection, vec!["test-a"])
            .await
            .unwrap();
        let output = Command::new(format!("{}/usr/bin/test-a", root)).output().expect("Running test-a executable failed; apparently installing the test-a package didn't work!");
        assert!(output.status.success());
        assert_eq!(
            String::from_utf8_lossy(&output.stdout),
            "hello from test-a-1.1\n"
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_get_package_details() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        server
            .add_packages(get_message().header(), &connection, vec!["test-a"])
            .await
            .unwrap();
        let mut builder_a = ApkPackagePartial::builder();
        builder_a
            .name("test-a".to_string())
            .version("1.0-r0".to_string())
            .description("Package A for apk-tools testsuite".to_string())
            .license("GPL".to_string())
            .url("http://alpinelinux.org".to_string())
            .installed_size(4096)
            .size(1233)
            .package_state(PackageState::Installed);
        let mut builder_e = ApkPackagePartial::builder();
        builder_e
            .name("test-e".to_string())
            .version("1.0-r0".to_string())
            .description("Package E for apk-tools testsuite".to_string())
            .license("GPL".to_string())
            .url("http://alpinelinux.org".to_string())
            .installed_size(4096)
            .size(1254)
            .package_state(PackageState::Available);

        let packages = server
            .get_packages_details(vec!["test-a", "test-e", "Unknown!"], DetailsFlags::all())
            .await
            .unwrap();
        assert_eq!(packages.len(), 3);
        assert_eq!(packages[0], builder_a.build());
        assert_eq!(packages[1], builder_e.build());
        assert!(packages[2].is_err());

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_delete_required_package() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        server
            .add_packages(get_message().header(), &connection, vec!["test-e"])
            .await
            .unwrap();
        let testa_exe_path = Path::new(&root).join("usr").join("bin").join("test-a");
        assert!(testa_exe_path.exists());
        let e = server
            .delete_packages(get_message().header(), &connection, vec!["test-a"])
            .await;
        assert!(e.is_err());
        assert_eq!(
            e.err().unwrap().to_string(),
            "dev.Cogitri.apkPolkit2.Anyhow: Package test-a is still required by the following packages: test-b test-e"
        );
        assert!(testa_exe_path.exists());

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_upgrade_subpackage() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo.clone()), Some(root.clone()));

        server
            .add_packages(
                get_message().header(),
                &connection,
                vec!["test-a-sub", "test-e"],
            )
            .await
            .unwrap();
        let output = Command::new(format!("{}/usr/bin/test-a", root)).output().expect("Running test-a executable failed; apparently installing the test-a package didn't work!");
        assert!(output.status.success());
        assert_eq!(output.stdout, b"hello from test-a-1.0\n");

        build_additional_package("test-a-new", &repo, &root);

        // No change should happen here
        server
            .upgrade_packages(get_message().header(), &connection, vec!["test-e"])
            .await
            .unwrap();
        let output = Command::new(format!("{}/usr/bin/test-a", root)).output().expect("Running test-a executable failed; apparently installing the test-a package didn't work!");
        assert!(output.status.success());
        assert_eq!(output.stdout, b"hello from test-a-1.0\n");

        // Upgrade the subpackage, that should also upgrade the main pkg
        server
            .upgrade_packages(
                get_message().header(),
                &connection,
                vec!["test-e", "test-a-sub"],
            )
            .await
            .unwrap();
        let output = Command::new(format!("{}/usr/bin/test-a", root)).output().expect("Running test-a executable failed; apparently installing the test-a package didn't work!");
        assert!(output.status.success());
        assert_eq!(
            String::from_utf8_lossy(&output.stdout),
            "hello from test-a-1.1\n"
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_replace_upgrade_package() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo.clone()), Some(root.clone()));

        server
            .add_packages(get_message().header(), &connection, vec!["test-a"])
            .await
            .unwrap();

        build_additional_package("test-c", &repo, &root);

        let upgrade_list = server
            .list_upgradable_packages(DetailsFlags::all())
            .await
            .unwrap();

        let mut builder_a = ApkPackagePartial::builder();
        builder_a
            .name("test-a".to_string())
            .version("1.0-r0".to_string())
            .description("Package A for apk-tools testsuite".to_string())
            .license("GPL".to_string())
            .url("http://alpinelinux.org".to_string())
            .installed_size(4096)
            .size(1229)
            .package_state(PackageState::PendingRemoval);

        let mut builder_c = ApkPackagePartial::builder();
        builder_c
            .name("test-c".to_string())
            .version("1.0-r1".to_string())
            .description("Package C for apk-tools testsuite".to_string())
            .license("GPL".to_string())
            .url("http://alpinelinux.org".to_string())
            .installed_size(4096)
            .size(1250)
            .package_state(PackageState::PendingInstall);

        assert_eq!(upgrade_list.len(), 2);
        assert_eq!(upgrade_list[0], builder_a.build());
        assert_eq!(upgrade_list[1], builder_c.build());

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_add_bad_package() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        let e = server
            .add_packages(get_message().header(), &connection, vec!["test-a 1.0"])
            .await
            .err();
        assert_eq!(
                e,
                Some(crate::error::Error::Anyhow("test-a 1.0 is not a correctly formated world dependency, the format should be: name(@tag)([<>~=]version)".to_string()).into())
            );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_delete_bad_package() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        let e = server
            .delete_packages(get_message().header(), &connection, vec!["test-a 1.0"])
            .await
            .err();
        assert_eq!(
                e,
                Some(crate::error::Error::Anyhow("test-a 1.0 is not a correctly formated world dependency, the format should be: name(@tag)([<>~=]version)".to_string()).into())
            );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_upgrade_bad_package() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        let e = server
            .upgrade_packages(get_message().header(), &connection, vec!["test-a 1.0"])
            .await
            .err();
        assert_eq!(
            e.unwrap().to_string(),
            "dev.Cogitri.apkPolkit2.Anyhow: test-a 1.0 is not a correctly formated world dependency, the format should be: name(@tag)([<>~=]version)",
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_get_repositories_nonexistent_file() {
        let (root, repo, temp_dir) = start_fs_test();
        let server = DBusServer::new_explicit(Some(repo), Some(root.clone()));
        let e = server.list_repositories().await.err();
        assert_eq!(
            e.unwrap().to_string(),
            format!(
                "dev.Cogitri.apkPolkit2.Anyhow: Failed to get repositories from file {root}/etc/apk/repositories due to error No such file or directory (os error 2)"
            ),
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_update_repositories_nonexistent_repo() {
        let (root, _, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server =
            DBusServer::new_explicit(Some("THISFILEDOESNTEXIST".to_string()), Some(root.clone()));
        let e = server
            .update_repositories(get_message().header(), &connection)
            .await
            .err();
        assert_eq!(
            e.unwrap().to_string(),
            "dev.Cogitri.apkPolkit2.Anyhow: No such file or directory"
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_database_open_nonexistent_dir() {
        let server = DBusServer::new_explicit(None, Some("THISFILEDOESNTEXIST".to_string()));
        let e = server
            .list_upgradable_packages(DetailsFlags::all())
            .await
            .err();
        assert_eq!(
            e.unwrap().to_string(),
            "dev.Cogitri.apkPolkit2.Anyhow: Bad file descriptor"
        );
    }

    #[tokio::test]
    async fn test_get_package_details_not_installed_pkg() {
        let (root, repo, temp_dir) = start_fs_test();
        let server = DBusServer::new_explicit(Some(repo), Some(root.clone()));
        let e = server
            .get_packages_details(vec!["unknown_package"], DetailsFlags::all())
            .await
            .unwrap()[0]
            .err();
        assert_eq!(
            e,
            "Couldn\'t retrieve package details for package unknown_package".to_string()
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_search_file_owner_nonexistent_file() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));

        server
            .add_packages(get_message().header(), &connection, vec!["test-e"])
            .await
            .unwrap();
        assert_eq!(
            server
                .search_files_owners(vec!["/usr/bin/test"], DetailsFlags::all())
                .await
                .unwrap()[0],
            ApkPackagePartial::builder().build()
        );

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_change_repository() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));
        let message = get_message();
        let url = "https://example.org";

        let repo_file_path = Path::new(&root)
            .join("etc")
            .join("apk")
            .join("repositories");
        create_dir_all(repo_file_path.parent().unwrap())
            .await
            .unwrap();
        let _repo_file = File::create(&repo_file_path).await.unwrap();

        server
            .add_repository(message.header().clone(), &connection, url)
            .await
            .unwrap();
        let mut repo = ApkRepository::new(
            url,
            repo_file_path.as_os_str().to_str().map(|s| s.to_string()),
        );
        assert_eq!(server.list_repositories().await.unwrap()[0].0, repo);

        server
            .change_repository(
                message.header().clone(),
                &connection,
                url,
                ChangeRepositoryActions::Disable as i32,
            )
            .await
            .unwrap();
        repo.enabled = false;
        assert_eq!(server.list_repositories().await.unwrap()[0].0, repo);

        server
            .change_repository(
                message.header().clone(),
                &connection,
                url,
                ChangeRepositoryActions::Enable as i32,
            )
            .await
            .unwrap();
        repo.enabled = true;
        assert_eq!(server.list_repositories().await.unwrap()[0].0, repo);

        temp_dir.close().unwrap();
    }

    fn assert_pkg_exists_in_cache(root: String, pkg_name: &str) {
        assert!(
            read_dir(format!("{}/var/cache/apk", root))
                .unwrap()
                .any(|file| {
                    file.as_ref()
                        .unwrap()
                        .path()
                        .file_name()
                        .unwrap()
                        .to_str()
                        .unwrap()
                        .starts_with(pkg_name)
                })
        );
    }

    #[tokio::test]
    async fn test_cache() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo), Some(root.clone()));
        let message = get_message();

        server
            .cache_packages(message.header(), &connection, vec!["test-a"], "", false)
            .await
            .unwrap();
        assert_pkg_exists_in_cache(root, "test-a");

        temp_dir.close().unwrap();
    }

    #[tokio::test]
    async fn test_cache_repos_file() {
        let (root, repo, temp_dir) = start_fs_test();
        let connection = Connection::session().await.unwrap();
        let mut server = DBusServer::new_explicit(Some(repo.clone()), Some(root.clone()));
        let message = get_message();
        std::fs::create_dir_all(format!("{root}/tmp")).unwrap();
        let mut repos_file = File::create(format!("{}/tmp/repositories", &root))
            .await
            .unwrap();
        let line = &repo;
        repos_file.write_all(line.as_bytes()).await.unwrap();

        server
            .cache_packages(
                message.header(),
                &connection,
                vec!["test-a"],
                &format!("{}/tmp/repositories", &root),
                false,
            )
            .await
            .unwrap();
        assert_pkg_exists_in_cache(root, "test-a");

        temp_dir.close().unwrap();
    }
}
