/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use apk_tools_sys::apk_blob_t;
use std::ffi::{CString, c_long};
use std::fmt::{Display, Formatter};
use std::os::raw::c_char;

/**
 * Wrapper around apk_blob_t. apk_blob_t's ptr is sometimes modified (pointer algorithmic), so we save the original pointer here and free it on drop.
 */
pub(crate) struct ApkBlob {
    blob: apk_blob_t,
    ptr: Option<*mut c_char>,
}

impl ApkBlob {
    pub fn new(string: &str) -> Self {
        let cstring = CString::new(string).unwrap();
        let blob = apk_blob_t {
            len: c_long::try_from(cstring.as_bytes().len()).unwrap(),
            ptr: cstring.into_raw(),
        };

        Self {
            ptr: Some(blob.ptr),
            blob,
        }
    }

    pub fn from_ptr(ptr: *mut apk_blob_t) -> Option<Self> {
        let raw = unsafe { ptr.as_mut() }?;
        Some(Self {
            ptr: None,
            blob: *raw,
        })
    }

    #[cfg(test)]
    pub fn from_value(value: apk_blob_t) -> Self {
        Self {
            ptr: None,
            blob: value,
        }
    }

    pub fn is_valid(&self) -> bool {
        //!self.blob.ptr.is_null() && self.blob.len < 0
        !(self.blob.ptr.is_null() || self.blob.len > 0)
    }

    pub(crate) fn as_mut(&mut self) -> &mut apk_blob_t {
        &mut self.blob
    }

    pub(crate) fn as_ref(&self) -> &apk_blob_t {
        &self.blob
    }
}

impl Display for ApkBlob {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let slice = unsafe {
            std::slice::from_raw_parts::<u8>(
                self.blob.ptr.cast(),
                usize::try_from(self.blob.len).unwrap(),
            )
        };
        let str = CString::new(slice).unwrap().into_string().unwrap();
        f.write_str(&str)
    }
}

impl Drop for ApkBlob {
    fn drop(&mut self) {
        unsafe {
            if let Some(ptr) = self.ptr {
                drop(CString::from_raw(ptr));
            }
        }
    }
}

#[cfg(test)]
mod test {
    use crate::{apk_blob::ApkBlob, global_allocator::GlobalAllocator};

    #[test]
    fn test_apk_blob_display() {
        let expected = "hello world!";
        let blob = ApkBlob::new(expected);
        assert_eq!(blob.to_string(), expected);
    }

    #[test]
    fn test_apk_blob_from_ptr_display() {
        let expected = "hello world!";
        let raw = GlobalAllocator::get_apk_blob(expected);
        let blob = ApkBlob::from_ptr(raw).unwrap();
        assert_eq!(blob.to_string(), expected);
    }
}
