/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::ApkChangeArray;
use apk_tools_sys::apk_changeset;

pub struct ApkChangeset {
    inner: Box<apk_changeset>,
    owns_changes: bool,
}

impl ApkChangeset {
    pub fn mark_new_packages(&mut self) {
        assert!(!self.inner.changes.is_null());

        let change_array = ApkChangeArray::from_ptr(self.inner.changes).unwrap();
        for change in change_array.iter() {
            if let Some(mut new_pkg) = change.new_pkg() {
                new_pkg.set_marked(true);
            }
        }
    }

    pub fn changes(&self) -> Option<ApkChangeArray> {
        ApkChangeArray::from_ptr(self.inner.changes)
    }

    pub(crate) fn from_raw(raw: apk_changeset, owns_changes: bool) -> Self {
        Self {
            inner: Box::new(raw),
            owns_changes,
        }
    }

    pub(crate) fn as_mut(&mut self) -> &mut apk_changeset {
        self.inner.as_mut()
    }
}

impl Drop for ApkChangeset {
    fn drop(&mut self) {
        if self.owns_changes {
            drop(ApkChangeArray::from_ptr_owned(self.inner.changes))
        }
    }
}

#[cfg(test)]
mod test {
    use crate::ApkChangeset;
    use crate::apk_change_array::ApkChangeArray;
    use apk_tools_sys::{apk_change, apk_changeset, apk_package};

    #[test]
    fn test_mark_new_packages() {
        let mut pkg1 = apk_package::default();
        let mut pkg2 = apk_package::default();
        let mut change1 = apk_change::default();
        change1.new_pkg = &mut pkg1;
        let mut change2 = apk_change::default();
        change2.new_pkg = &mut pkg2;
        let change3 = apk_change::default();

        let mut changes = ApkChangeArray::new();
        changes.push(change1);
        changes.push(change2);
        changes.push(change3);

        let mut raw_changeset = apk_changeset::default();
        raw_changeset.changes = changes.as_mut();
        let mut changeset = ApkChangeset::from_raw(raw_changeset, false);

        changeset.mark_new_packages();

        let marked_changes = ApkChangeArray::from_ptr(changeset.inner.changes).unwrap();
        unsafe {
            assert_eq!((*marked_changes.as_slice()[0].new_pkg).marked(), 1);
        }
        unsafe {
            assert_eq!((*marked_changes.as_slice()[1].new_pkg).marked(), 1);
        }
        assert!(marked_changes.as_slice()[2].new_pkg.is_null());
    }
}
