/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::ApkStringArray;
use apk_tools_sys::{
    APK_OPENF_CACHE_WRITE, APK_OPENF_CREATE, APK_OPENF_NO_AUTOUPDATE, APK_OPENF_READ,
    APK_OPENF_USERMODE, APK_OPENF_WRITE, apk_ctx, apk_ctx_free, apk_ctx_init, apk_ctx_prepare,
};
use std::cell::OnceCell;
use std::{ffi::CString, os::raw::c_uint};

#[derive(Default, Debug)]
pub struct ApkCtxBuilder {
    additional_repositories: ApkStringArray,
    allow_untrusted: bool,
    read_only: bool,
    repositories_file: Option<String>,
    root: Option<String>,
    usermode: bool,
}

impl ApkCtxBuilder {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn add_additional_repository(&mut self, value: &str) -> &mut Self {
        self.additional_repositories.push(value);
        self
    }

    pub fn set_allow_untrusted(&mut self, value: bool) -> &mut Self {
        self.allow_untrusted = value;
        self
    }

    pub fn set_read_only(&mut self, value: bool) -> &mut Self {
        self.read_only = value;
        self
    }

    pub fn set_repositories_file(&mut self, value: String) -> &mut Self {
        self.repositories_file = Some(value);
        self
    }

    pub fn set_root(&mut self, value: String) -> &mut Self {
        self.root = Some(value);
        self
    }

    pub fn set_usermode(&mut self, value: bool) -> &mut Self {
        self.usermode = value;
        self
    }

    pub fn build(self) -> ApkCtx {
        let mut ctx = Box::<apk_ctx>::default();

        unsafe {
            apk_ctx_init(ctx.as_mut());
        }

        ctx.lock_wait = 1;

        if self.allow_untrusted {
            // FIXME: bindgen doesn't understand the BIT(x) macro apk uses in apk_context.h
            const APK_ALLOW_UNTRUSTED: u32 = 1 << 3;
            ctx.flags = APK_ALLOW_UNTRUSTED;
        }

        if self.read_only {
            ctx.open_flags = c_uint::from(APK_OPENF_READ | APK_OPENF_NO_AUTOUPDATE);
        } else {
            ctx.open_flags = c_uint::from(
                APK_OPENF_READ
                    | APK_OPENF_WRITE
                    | APK_OPENF_NO_AUTOUPDATE
                    | APK_OPENF_CACHE_WRITE
                    | APK_OPENF_CREATE,
            );
        }

        if self.usermode {
            ctx.open_flags |= c_uint::from(APK_OPENF_USERMODE);
        }

        #[cfg(test)]
        {
            ctx.open_flags |= c_uint::from(apk_tools_sys::APK_OPENF_NO_CMDLINE_REPOS);
        }

        let mut ctx = ApkCtx::from_value(ctx);

        if !self.additional_repositories.is_empty() {
            ctx.set_additional_repositories(self.additional_repositories);
        }

        if let Some(root) = self.root {
            ctx.set_root(CString::new(root).unwrap());
        }

        if let Some(repositories_file) = self.repositories_file {
            ctx.set_repositories_file(CString::new(repositories_file).unwrap());
        }

        ctx.prepare();

        ctx
    }
}

#[derive(Debug)]
pub struct ApkCtx {
    inner: Box<apk_ctx>,
    root: OnceCell<CString>,
    repositories_file: OnceCell<CString>,
}

impl ApkCtx {
    pub fn builder() -> ApkCtxBuilder {
        ApkCtxBuilder::new()
    }

    pub(crate) fn as_mut(&mut self) -> &mut apk_ctx {
        self.inner.as_mut()
    }

    fn from_value(inner: Box<apk_ctx>) -> Self {
        Self {
            inner,
            root: OnceCell::new(),
            repositories_file: OnceCell::new(),
        }
    }

    fn set_additional_repositories(&mut self, val: ApkStringArray) {
        drop(ApkStringArray::from_ptr_owned(self.inner.repository_list));
        self.inner.repository_list = val.steal();
    }

    fn set_root(&mut self, val: CString) {
        self.root.set(val).unwrap();
        self.inner.root = self.root.get().unwrap().as_ptr();
    }

    fn set_repositories_file(&mut self, val: CString) {
        self.repositories_file.set(val).unwrap();
        self.inner.repositories_file = self.repositories_file.get().unwrap().as_ptr();
    }

    fn prepare(&mut self) {
        unsafe {
            apk_ctx_prepare(self.inner.as_mut());
        }
    }
}

impl Drop for ApkCtx {
    fn drop(&mut self) {
        if let Some(mut repo_list) = ApkStringArray::from_ptr(self.inner.repository_list) {
            repo_list.clear();
        }

        unsafe { apk_ctx_free(self.inner.as_mut()) };
    }
}

#[cfg(test)]
mod test {
    use super::ApkCtx;
    use apk_tools_sys::APK_OPENF_WRITE;
    use std::ffi::{CStr, c_uint};

    #[test]
    fn test_set_root() {
        let root = "/test";
        let mut builder = ApkCtx::builder();
        builder.set_root(root.to_string());
        let options = builder.build();
        let c_root = unsafe { CStr::from_ptr(options.inner.root) };
        assert_eq!(c_root.to_str().unwrap(), root);
    }

    #[test]
    fn test_repositories_file() {
        let file = "/test";
        let mut builder = ApkCtx::builder();
        builder.set_repositories_file(file.to_string());
        let options = builder.build();
        let c_file = unsafe { CStr::from_ptr(options.inner.repositories_file) };
        assert_eq!(c_file.to_str().unwrap(), file);
    }

    #[test]
    fn test_readonly() {
        let mut builder = ApkCtx::builder();
        builder.set_read_only(true);
        let options = builder.build();
        assert_eq!(options.inner.open_flags & c_uint::from(APK_OPENF_WRITE), 0);
    }
}
