/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::{
    ApkBlob, ApkChangeset, ApkCtx, ApkDependencyArray, ApkName, ApkPackage, ApkRepository,
    providers::{ApkDatabaseProvider, ApkDatabaseProviderImpl},
};
use anyhow::Result;
use apk_tools_sys::{
    APK_SOLVERF_AVAILABLE, APK_SOLVERF_UPGRADE, apk_crypto_init, apk_dependency, apk_name,
};
use std::{cell::OnceCell, collections::HashSet, fmt::Debug};

#[derive(Debug)]
pub struct ApkDatabase {
    provider: Box<dyn ApkDatabaseProvider>,
    apk_ctx: OnceCell<ApkCtx>,
}

impl ApkDatabase {
    pub fn new() -> Self {
        let provider = Box::new(ApkDatabaseProviderImpl::new());
        Self::from_provider(provider)
    }

    pub fn open(&mut self, mut ctx: ApkCtx) -> Result<()> {
        unsafe {
            apk_crypto_init();
        }
        self.provider.apk_db_init(ctx.as_mut());
        let ret = self.provider.apk_db_open();
        self.apk_ctx.set(ctx).unwrap();
        ret
    }

    pub fn commit(&mut self, new_world: &mut ApkDependencyArray) -> Result<()> {
        self.provider.apk_solver_commit(0, new_world.as_mut())
    }

    pub fn commit_changeset(
        &mut self,
        changeset: &mut ApkChangeset,
        world: &mut ApkDependencyArray,
    ) -> Result<()> {
        self.provider
            .apk_solver_commit_changeset(changeset.as_mut(), world.as_mut())
    }

    pub fn solve_upgrade_changeset(
        &mut self,
        new_world: &mut ApkDependencyArray,
    ) -> Result<ApkChangeset> {
        let raw = self.provider.apk_solver_solve(
            u8::try_from(APK_SOLVERF_UPGRADE | APK_SOLVERF_AVAILABLE)?,
            new_world.as_mut(),
        )?;
        Ok(ApkChangeset::from_raw(raw, true))
    }

    pub fn solve_delete_changeset(
        &mut self,
        new_world: &mut ApkDependencyArray,
    ) -> Result<ApkChangeset> {
        let raw = self.provider.apk_solver_solve(0, new_world.as_mut())?;
        Ok(ApkChangeset::from_raw(raw, true))
    }

    pub fn get_reverse_dependencies(&mut self, pkgname: &str) -> Option<HashSet<String>> {
        let name = self.get_apk_name(pkgname)?;

        fn add(pkg: &ApkPackage, pkgnames: &mut HashSet<String>) {
            let pkg_name = pkg.name();
            pkgnames.insert(pkg_name.name().to_string());

            for reverse_dependency in pkg_name.rdepends().unwrap().iter() {
                let providers = reverse_dependency.providers().unwrap();
                for pkg in providers.iter().filter_map(|p| p.pkg()) {
                    if pkg.has_installed_package() || pkg.is_marked() {
                        add(&pkg, pkgnames);
                    }
                }
            }
        }

        let mut result = HashSet::new();
        if let Some(providers) = name.providers() {
            for provider in providers.iter() {
                if let Some(pkg) = provider.pkg() {
                    if pkg.is_marked() {
                        add(&pkg, &mut result);
                    }
                }
            }
        }

        result.remove(pkgname);

        if result.is_empty() {
            None
        } else {
            Some(result)
        }
    }

    pub fn get_world(&self) -> ApkDependencyArray {
        ApkDependencyArray::from_ptr(self.provider.get_world())
            .unwrap()
            .clone()
    }

    pub fn get_file_owner(&mut self, filename: &str) -> Option<ApkPackage> {
        let blob = ApkBlob::new(filename);
        let pkg = self.provider.apk_db_get_file_owner(*blob.as_ref());
        ApkPackage::from_ptr(pkg)
    }

    pub fn update_repositories(&mut self) -> Result<()> {
        let repos = self.provider.get_repos();
        for x in repos {
            self.provider.apk_cache_download_repo(x)?;
        }

        Ok(())
    }

    pub fn query_name(&mut self, pkgname: &str) -> Option<ApkName> {
        let blob = ApkBlob::new(pkgname);
        let name = self.provider.apk_db_query_name(*blob.as_ref());
        ApkName::from_ptr(name)
    }

    pub fn check_world(&mut self) -> Result<()> {
        let mut world = ApkDependencyArray::from_ptr(self.provider.get_world()).unwrap();
        self.provider.apk_db_check_world(world.as_mut())
    }

    pub fn select_repo(&mut self, pkg: &mut ApkPackage) -> Option<ApkRepository> {
        let repo = self.provider.apk_db_select_repo(pkg.as_mut());
        ApkRepository::from_ptr(repo)
    }

    pub fn cache_package(
        &mut self,
        repo: &mut ApkRepository,
        package: &mut ApkPackage,
    ) -> Result<()> {
        self.provider
            .apk_cache_download_pkg(repo.as_mut(), package.as_mut())
    }

    pub(crate) fn pull_dependency(&mut self, pkgname: &mut ApkBlob) -> apk_dependency {
        self.provider.apk_blob_pull_dep(pkgname.as_mut())
    }

    fn from_provider(provider: Box<dyn ApkDatabaseProvider>) -> Self {
        Self {
            provider,
            apk_ctx: OnceCell::new(),
        }
    }

    fn get_apk_name(&mut self, pkgname: &str) -> Option<ApkName> {
        let key = ApkBlob::new(pkgname);
        let hash_item = self.provider.apk_hash_get_hashed(*key.as_ref());
        let raw_name: *mut apk_name = hash_item.cast();

        ApkName::from_ptr(raw_name)
    }
}

#[cfg(test)]
mod test {
    use crate::apk_blob::ApkBlob;
    use crate::{
        ApkCtx, ApkDatabase, global_allocator::GlobalAllocator, providers::ApkDatabaseProvider,
    };
    use apk_tools_sys::{
        apk_blob_t, apk_changeset, apk_ctx, apk_dependency, apk_dependency_array, apk_hash_item,
        apk_name, apk_package, apk_repository,
    };
    use std::ffi::CStr;
    use std::fmt::Debug;

    #[derive(Debug, Default)]
    struct ApkProviderTest {
        is_opened: bool,
        is_init: bool,
        file_owner_pkgs: Vec<apk_package>,
    }

    impl ApkDatabaseProvider for ApkProviderTest {
        fn apk_blob_pull_dep(&mut self, _name: &mut apk_blob_t) -> apk_dependency {
            todo!()
        }

        fn apk_cache_download_pkg(
            &mut self,
            _repository: *mut apk_repository,
            _package: *mut apk_package,
        ) -> anyhow::Result<()> {
            todo!()
        }

        fn apk_cache_download_repo(&mut self, _repo: &mut apk_repository) -> anyhow::Result<()> {
            todo!()
        }

        fn apk_db_check_world(&mut self, _world: &mut apk_dependency_array) -> anyhow::Result<()> {
            todo!()
        }

        fn apk_db_get_file_owner(&mut self, _filename: apk_blob_t) -> *mut apk_package {
            assert!(self.is_init);
            assert!(self.is_opened);
            let mut pkg = apk_package::default();
            pkg.name = GlobalAllocator::get_apk_name("test");
            self.file_owner_pkgs.push(pkg);
            self.file_owner_pkgs.iter_mut().last().unwrap() as *mut apk_package
        }

        fn apk_db_init(&mut self, _ctx: &mut apk_ctx) {
            assert!(!self.is_init);
            assert!(!self.is_opened);
            self.is_init = true;
        }

        fn apk_db_open(&mut self) -> anyhow::Result<()> {
            assert!(self.is_init);
            assert!(!self.is_opened);
            self.is_opened = true;
            Ok(())
        }

        fn apk_db_query_name(&mut self, name: apk_blob_t) -> *mut apk_name {
            let blob = ApkBlob::from_value(name);
            match blob.to_string().as_str() {
                "" => std::ptr::null_mut(),
                _ => GlobalAllocator::get_apk_name(&blob.to_string()),
            }
        }

        fn apk_db_select_repo(&mut self, _pkg: *mut apk_package) -> *mut apk_repository {
            todo!()
        }

        fn apk_hash_get_hashed(&mut self, key: apk_blob_t) -> apk_hash_item {
            assert!(self.is_init);
            assert!(self.is_opened);
            let key_str = unsafe { CStr::from_ptr(key.ptr) }.to_str().unwrap();
            let name = GlobalAllocator::get_apk_name(key_str);
            name.cast()
        }

        fn apk_solver_commit(
            &mut self,
            _solver_flags: u8,
            _world: &mut apk_dependency_array,
        ) -> anyhow::Result<()> {
            assert!(self.is_init);
            assert!(self.is_opened);
            Ok(())
        }

        fn apk_solver_commit_changeset(
            &mut self,
            _changeset: &mut apk_changeset,
            _world: &mut apk_dependency_array,
        ) -> anyhow::Result<()> {
            assert!(self.is_init);
            assert!(self.is_opened);
            Ok(())
        }

        fn apk_solver_solve(
            &mut self,
            _solver_flags: u8,
            _world: &mut apk_dependency_array,
        ) -> anyhow::Result<apk_changeset> {
            assert!(self.is_init);
            assert!(self.is_opened);
            todo!()
        }

        fn get_world(&self) -> &mut apk_dependency_array {
            assert!(self.is_init);
            assert!(self.is_opened);
            todo!()
        }

        fn get_repos(&mut self) -> Vec<&'static mut apk_repository> {
            todo!()
        }
    }

    fn get_database() -> ApkDatabase {
        ApkDatabase::from_provider(Box::new(ApkProviderTest::default()))
    }

    fn get_apk_ctx() -> ApkCtx {
        ApkCtx::builder().build()
    }

    // fn get_provider(db: &ApkDatabase) -> &ApkProviderTest {
    //     ((&*db.provider) as &dyn Any)
    //         .downcast_ref::<ApkProviderTest>()
    //         .unwrap()
    // }

    #[test]
    fn test_get_apk_name() {
        let mut db = get_database();
        db.open(get_apk_ctx()).unwrap();
        let pkgname = "myname";
        let name = db.get_apk_name(pkgname);
        assert!(name.is_some());
        let result = name.unwrap();
        assert_eq!(result.name(), pkgname);
    }

    #[test]
    #[should_panic]
    fn test_get_apk_name_before_open() {
        let mut db = get_database();
        db.get_apk_name("test");
    }

    #[test]
    fn test_get_file_owner() {
        let mut db = get_database();
        db.open(get_apk_ctx()).unwrap();
        let owner = db.get_file_owner("file.txt");
        assert!(owner.is_some());
        assert_eq!(owner.unwrap().name().name(), "test");
    }

    #[test]
    fn test_query_name() {
        let mut db = get_database();
        db.open(get_apk_ctx()).unwrap();
        let name = db.query_name("test");
        assert!(name.is_some());
        assert_eq!(name.unwrap().name(), "test");
    }

    #[test]
    fn test_query_name_none() {
        let mut db = get_database();
        db.open(get_apk_ctx()).unwrap();
        let name = db.query_name("");
        assert!(name.is_none());
    }
}
