/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::{ApkBlob, ApkDatabase, ApkName, ApkProviderArray};
use anyhow::{Result, anyhow};
use apk_tools_sys::{APK_DEPMASK_ANY, APK_DEPMASK_CHECKSUM, apk_atom_null, apk_dependency};
use std::ptr::addr_of_mut;

#[derive(Clone, Debug)]
pub struct ApkDependency {
    inner: apk_dependency,
}

impl ApkDependency {
    pub fn from_pkgname(db: &mut ApkDatabase, pkgname: &str) -> Result<Option<Self>> {
        let mut blobbed_name = ApkBlob::new(pkgname);
        let dependency = db.pull_dependency(&mut blobbed_name);
        if blobbed_name.is_valid() {
            if unsafe {
                ApkProviderArray::from_ptr((*dependency.name).providers)
                    .unwrap()
                    .is_empty()
            } {
                Ok(None)
            } else {
                Ok(Some(Self::from_value(dependency)))
            }
        } else {
            Err(anyhow!(
                "{pkgname} is not a correctly formated world dependency, the format should be: name(@tag)([<>~=]version)"
            ))
        }
    }

    pub fn name(&self) -> ApkName {
        ApkName::from_ptr(self.inner.name).unwrap()
    }

    pub fn has_checksum_depmask(&self) -> bool {
        self.inner.op as u32 == APK_DEPMASK_CHECKSUM
    }

    pub fn reset_for_upgrade(&mut self) {
        self.inner.op = u8::try_from(APK_DEPMASK_ANY).unwrap();
        self.inner.version = addr_of_mut!(apk_atom_null);
    }

    pub(crate) fn into_inner(self) -> apk_dependency {
        self.inner
    }

    pub(crate) fn from_value(value: apk_dependency) -> Self {
        Self { inner: value }
    }
}

impl PartialEq<&apk_dependency> for &ApkDependency {
    fn eq(&self, other: &&apk_dependency) -> bool {
        other.name == self.inner.name
    }
}

#[cfg(test)]
mod test {
    use crate::ApkDependency;
    use crate::global_allocator::GlobalAllocator;
    use apk_tools_sys::{APK_DEPMASK_ANY, APK_DEPMASK_CHECKSUM, apk_atom_null, apk_dependency};
    use std::ptr::addr_of_mut;

    #[test]
    fn test_name_ptr_eq() {
        let mut raw_dep1 = apk_dependency::default();
        let mut raw_dep2 = apk_dependency::default();

        let name = GlobalAllocator::get_apk_name("test");

        raw_dep1.name = name;
        raw_dep2.name = name;

        assert_eq!(&ApkDependency::from_value(raw_dep1), &raw_dep2);
    }

    #[test]
    fn test_name_str_not_eq() {
        let mut raw_dep1 = apk_dependency::default();
        let mut raw_dep2 = apk_dependency::default();

        let name1 = GlobalAllocator::get_apk_name("test");
        let name2 = GlobalAllocator::get_apk_name("test");

        raw_dep1.name = name1;
        raw_dep2.name = name2;

        assert_ne!(&ApkDependency::from_value(raw_dep1), &raw_dep2);
    }

    #[test]
    fn test_has_checksum_depmask() {
        let mut dep = ApkDependency::from_value(apk_dependency::default());
        assert!(!dep.has_checksum_depmask());
        dep.inner.op = APK_DEPMASK_CHECKSUM as u8;
        assert!(dep.has_checksum_depmask());
    }

    #[test]
    fn test_reset_for_upgrade() {
        let mut dep = ApkDependency::from_value(apk_dependency::default());
        dep.inner.version = GlobalAllocator::get_apk_blob("test");
        assert_eq!(dep.inner.op, 0);
        assert_ne!(dep.inner.version, addr_of_mut!(apk_atom_null));

        dep.reset_for_upgrade();

        assert_eq!(dep.inner.op, APK_DEPMASK_ANY as u8);
        assert_eq!(dep.inner.version, addr_of_mut!(apk_atom_null));
    }
}
