/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::{ApkNameArray, ApkPackage, ApkProviderArray, generate_apk_base_impl};
use apk_tools_sys::{
    APK_SOLVERF_AVAILABLE, APK_SOLVERF_UPGRADE, apk_name, apk_pkg_get_installed,
    apk_solver_set_name_flags,
};
use std::ffi::CStr;
use std::fmt::{Display, Formatter};

pub struct ApkName {
    inner: &'static mut apk_name,
}

generate_apk_base_impl!(ApkName, apk_name);

impl ApkName {
    pub fn name(&self) -> &str {
        unsafe { CStr::from_ptr(self.inner.name.as_ptr()).to_str().unwrap() }
    }

    pub fn providers(&self) -> Option<ApkProviderArray> {
        ApkProviderArray::from_ptr(self.inner.providers)
    }

    pub fn rdepends(&self) -> Option<ApkNameArray> {
        ApkNameArray::from_ptr(self.inner.rdepends)
    }

    pub fn set_update_flags(&mut self) {
        const APK_SOLVERF_NO_FLAGS: u16 = 0;
        unsafe {
            apk_solver_set_name_flags(
                self.inner,
                u16::try_from(APK_SOLVERF_UPGRADE | APK_SOLVERF_AVAILABLE).unwrap(),
                APK_SOLVERF_NO_FLAGS,
            );
        }
    }

    pub fn get_installed_package(&mut self) -> Option<ApkPackage> {
        let pkg = unsafe { apk_pkg_get_installed(self.inner) };
        ApkPackage::from_ptr(pkg)
    }
}

impl Display for ApkName {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        Display::fmt(self.name(), f)
    }
}

#[cfg(test)]
mod test {
    use crate::ApkName;
    use crate::global_allocator::GlobalAllocator;

    #[test]
    fn test_name() {
        let expected = "test";
        let raw = GlobalAllocator::get_apk_name(expected);
        let name = ApkName::from_ptr(raw).unwrap();
        assert_eq!(name.name(), expected);
    }
}
