/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::apk_blob::ApkBlob;
use crate::{ApkName, generate_apk_base_impl};
use apk_tools_sys::{
    APK_VERSION_EQUAL, APK_VERSION_GREATER, APK_VERSION_LESS, apk_package, apk_pkg_version_compare,
};
use std::ffi::c_uchar;

#[derive(Debug)]
pub struct ApkPackage {
    inner: &'static mut apk_package,
}

generate_apk_base_impl!(ApkPackage, apk_package);

impl ApkPackage {
    pub fn is_marked(&self) -> bool {
        self.inner.marked() == 1
    }

    pub fn set_marked(&mut self, marked: bool) {
        self.inner.set_marked(marked as i32 as c_uchar);
    }

    pub fn has_installed_package(&self) -> bool {
        !self.inner.ipkg.is_null()
    }

    pub fn name(&self) -> ApkName {
        ApkName::from_ptr(self.inner.name).unwrap()
    }

    pub fn version(&self) -> String {
        ApkBlob::from_ptr(self.inner.version).unwrap().to_string()
    }

    pub fn license(&self) -> String {
        ApkBlob::from_ptr(self.inner.license).unwrap().to_string()
    }

    pub fn url(&self) -> Option<String> {
        ApkBlob::from_ptr(self.inner.url).map(|b| b.to_string())
    }

    pub fn description(&self) -> Option<String> {
        ApkBlob::from_ptr(self.inner.description).map(|b| b.to_string())
    }

    pub fn size(&self) -> u64 {
        self.inner.size
    }

    pub fn installed_size(&self) -> u64 {
        self.inner.installed_size
    }

    pub fn origin(&self) -> Option<String> {
        let origin = unsafe { self.inner.origin.as_mut() }?;
        Some(ApkBlob::from_ptr(origin)?.to_string())
    }

    pub fn ptr_compare(&self, other: &Self) -> bool {
        std::ptr::eq(self.inner, other.inner)
    }
}

impl PartialEq for ApkPackage {
    fn eq(&self, other: &Self) -> bool {
        self.name().name() == other.name().name() && self.version() == other.version()
    }
}

impl PartialOrd for ApkPackage {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        let result =
            u32::try_from(unsafe { apk_pkg_version_compare(self.inner, other.inner) }).unwrap();
        if result & APK_VERSION_LESS != 0 {
            Some(std::cmp::Ordering::Less)
        } else if result & APK_VERSION_EQUAL != 0 {
            Some(std::cmp::Ordering::Equal)
        } else if result & APK_VERSION_GREATER != 0 {
            Some(std::cmp::Ordering::Greater)
        } else {
            None
        }
    }
}

#[cfg(test)]
mod test {
    use crate::ApkPackage;
    use crate::global_allocator::GlobalAllocator;

    #[test]
    fn test_is_marked() {
        let raw = GlobalAllocator::get_apk_package();
        let pkg = ApkPackage::from_ptr(raw).unwrap();
        assert!(!pkg.is_marked());

        pkg.inner.set_marked(1);

        assert!(pkg.is_marked());
    }

    #[test]
    fn test_set_marked() {
        let raw = GlobalAllocator::get_apk_package();
        let mut pkg = ApkPackage::from_ptr(raw).unwrap();
        assert!(!pkg.is_marked());

        pkg.set_marked(true);

        assert!(pkg.is_marked());
    }

    #[test]
    fn test_version() {
        let raw = GlobalAllocator::get_apk_package();
        let pkg = ApkPackage::from_ptr(raw).unwrap();
        let expected = "1.0.0";

        pkg.inner.version = GlobalAllocator::get_apk_blob(expected);

        assert_eq!(pkg.version(), expected);
    }

    #[test]
    fn test_license() {
        let raw = GlobalAllocator::get_apk_package();
        let pkg = ApkPackage::from_ptr(raw).unwrap();
        let expected = "GPL-3.0-or-later";

        pkg.inner.license = GlobalAllocator::get_apk_blob(expected);

        assert_eq!(pkg.license(), expected);
    }

    #[test]
    fn test_url() {
        let raw = GlobalAllocator::get_apk_package();
        let pkg = ApkPackage::from_ptr(raw).unwrap();
        let expected = "https://alpinelinux.org";
        assert!(pkg.url().is_none());

        pkg.inner.url = GlobalAllocator::get_apk_blob(expected);

        assert_eq!(pkg.url().unwrap(), expected);
    }

    #[test]
    fn test_description() {
        let raw = GlobalAllocator::get_apk_package();
        let pkg = ApkPackage::from_ptr(raw).unwrap();
        let expected = "https://alpinelinux.org";
        assert!(pkg.description().is_none());

        pkg.inner.description = GlobalAllocator::get_apk_blob(expected);

        assert_eq!(pkg.description().unwrap(), expected);
    }

    #[test]
    fn test_origin() {
        let raw = GlobalAllocator::get_apk_package();
        let pkg = ApkPackage::from_ptr(raw).unwrap();
        let expected = "test-a";
        assert!(pkg.origin().is_none());

        pkg.inner.origin = GlobalAllocator::get_apk_blob(expected);

        assert_eq!(pkg.origin().unwrap(), expected);
    }

    #[test]
    fn test_eq() {
        let raw1 = GlobalAllocator::get_apk_package();
        let pkg1 = ApkPackage::from_ptr(raw1).unwrap();
        let raw2 = GlobalAllocator::get_apk_package();
        let pkg2 = ApkPackage::from_ptr(raw2).unwrap();

        let name = "test-a";
        let version = "1.0.0";

        pkg1.inner.name = GlobalAllocator::get_apk_name(name);
        pkg2.inner.name = GlobalAllocator::get_apk_name(name);
        pkg1.inner.version = GlobalAllocator::get_apk_blob(version);
        pkg2.inner.version = GlobalAllocator::get_apk_blob(version);
        assert_eq!(pkg1, pkg2);

        pkg1.inner.version = GlobalAllocator::get_apk_blob(&format!("{version}1"));
        assert_ne!(pkg1, pkg2);
    }

    #[test]
    fn test_ord() {
        let raw1 = GlobalAllocator::get_apk_package();
        let pkg1 = ApkPackage::from_ptr(raw1).unwrap();
        let raw2 = GlobalAllocator::get_apk_package();
        let pkg2 = ApkPackage::from_ptr(raw2).unwrap();

        pkg1.inner.version = GlobalAllocator::get_apk_blob("1.0.0");
        pkg2.inner.version = GlobalAllocator::get_apk_blob("1.0.1");

        assert!(pkg1 < pkg2);
        assert!(pkg2 > pkg1);
    }
}
