/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::apk_change::ApkChange;
use crate::{generate_apk_array_drop_impl, generate_apk_array_impl};
use apk_tools_sys::{apk_change, apk_change_array};

#[derive(Debug)]
pub struct ApkChangeArray {
    inner: &'static mut apk_change_array,
    is_owned: bool,
}

generate_apk_array_impl!(ApkChangeArray, apk_change_array, apk_change);
generate_apk_array_drop_impl!(ApkChangeArray);

impl ApkChangeArray {
    pub fn iter(&self) -> impl Iterator<Item = ApkChange> + use<'_> {
        self.as_slice()
            .iter()
            .map(|&change| ApkChange::from_raw(change))
    }

    pub fn push(&mut self, change: apk_change) {
        unsafe {
            self.push_raw(change);
        }
    }
}

#[cfg(test)]
mod test {
    use super::ApkChangeArray;
    use crate::global_allocator::GlobalAllocator;
    use apk_tools_sys::apk_change;

    #[test]
    fn test_arr_as_slice() {
        let raw = ApkChangeArray::new_raw_array();
        let mut arr = ApkChangeArray::from_ptr_owned(raw).unwrap();
        assert_eq!(arr.as_slice().len(), 0);
        let change = apk_change::default();
        unsafe { arr.push_raw(change) };
        assert_eq!(arr.as_slice().len(), 1);
    }

    #[test]
    fn test_arr_copy() {
        let raw = ApkChangeArray::new_raw_array();
        let mut arr = ApkChangeArray::from_ptr_owned(raw).unwrap();
        let mut change = apk_change::default();
        change.new_pkg = GlobalAllocator::get_apk_package();
        unsafe { arr.push_raw(change) };
        assert_eq!(arr.as_slice().len(), 1);

        let arr_copy = arr.clone();
        assert_eq!(arr_copy.as_slice().len(), 1);
        assert!(std::ptr::eq(
            arr_copy.as_slice().last().unwrap().new_pkg,
            arr.as_slice().last().unwrap().new_pkg
        ));
    }
}
