/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::apk_dependency::ApkDependency;
use crate::{generate_apk_array_drop_impl, generate_apk_array_impl};
use apk_tools_sys::{apk_dependency, apk_dependency_array};

#[derive(Debug)]
pub struct ApkDependencyArray {
    inner: &'static mut apk_dependency_array,
    is_owned: bool,
}

generate_apk_array_impl!(ApkDependencyArray, apk_dependency_array, apk_dependency);
generate_apk_array_drop_impl!(ApkDependencyArray);
impl ApkDependencyArray {
    pub fn push(&mut self, dep: ApkDependency) {
        if !self.contains(&dep) {
            unsafe {
                let mut inner = self.inner as *mut _;
                Self::add(&mut inner, dep.into_inner());
                self.inner = inner.as_mut().unwrap();
            }
        }
    }

    pub fn delete(&mut self, dep: &ApkDependency) {
        let raw_deps = self.inner.as_mut_slice().unwrap();

        if let Some(pos) = raw_deps.iter().position(|x| dep == x) {
            raw_deps[pos] = *raw_deps.last().unwrap();
            self.inner.hdr.num -= 1;
        }
    }

    pub fn contains(&self, dependency: &ApkDependency) -> bool {
        for x in self.inner.as_slice().unwrap() {
            if dependency == x {
                return true;
            }
        }

        false
    }

    pub fn iter(&self) -> impl Iterator<Item = ApkDependency> + use<'_> {
        self.as_slice()
            .iter()
            .map(|&dependency| ApkDependency::from_value(dependency))
    }
}

#[cfg(test)]
mod test {
    use crate::global_allocator::GlobalAllocator;
    use crate::{ApkDependency, ApkDependencyArray};
    use apk_tools_sys::apk_dependency;

    fn get_dependency(pkgname: &str) -> ApkDependency {
        let mut dep = apk_dependency::default();
        dep.name = GlobalAllocator::get_apk_name(pkgname);
        ApkDependency::from_value(dep)
    }

    #[test]
    fn apk_dependency_array_new_is_not_null() {
        let _result = ApkDependencyArray::new();
    }

    #[test]
    fn apk_dependency_array_can_push_dependency() {
        let mut array = ApkDependencyArray::new();
        let dependency = get_dependency("test");
        array.push(dependency);
    }

    #[test]
    fn apk_dependency_array_can_check_contains_dependency() {
        let mut array = ApkDependencyArray::new();
        let dependency_test1 = get_dependency("test");
        let dependency_test2 = get_dependency("test2");
        array.push(dependency_test1.clone());
        assert!(array.contains(&dependency_test1));
        assert!(!array.contains(&dependency_test2));
    }

    #[test]
    fn apk_dependency_array_can_delete_dependency() {
        let mut array = ApkDependencyArray::new();
        let dependency_test1 = get_dependency("test");
        let dependency_test2 = get_dependency("test2");
        let dependency_test3 = get_dependency("test3");
        array.push(dependency_test1.clone());
        array.push(dependency_test2.clone());
        array.push(dependency_test3.clone());
        assert!(array.contains(&dependency_test1));
        assert!(array.contains(&dependency_test2));
        assert!(array.contains(&dependency_test3));

        array.delete(&dependency_test1);
        assert!(!array.contains(&dependency_test1));
        assert!(array.contains(&dependency_test2));
        assert!(array.contains(&dependency_test3));
    }

    #[test]
    fn apk_dependency_array_can_delete_last_dependency() {
        let mut array = ApkDependencyArray::new();
        let dependency = get_dependency("test");
        array.push(dependency.clone());
        assert!(array.contains(&dependency));
        array.delete(&dependency);
        assert!(!array.contains(&dependency));
    }

    #[test]
    fn test_arr_as_slice() {
        let raw = ApkDependencyArray::new_raw_array();
        let mut arr = ApkDependencyArray::from_ptr_owned(raw).unwrap();
        assert_eq!(arr.as_slice().len(), 0);
        let change = apk_dependency::default();
        unsafe { arr.push_raw(change) };
        assert_eq!(arr.as_slice().len(), 1);
    }

    #[test]
    fn test_arr_copy() {
        let raw = ApkDependencyArray::new_raw_array();
        let mut arr = ApkDependencyArray::from_ptr_owned(raw).unwrap();
        let mut dependency = apk_dependency::default();
        let name = GlobalAllocator::get_apk_name("test");
        dependency.name = name;
        unsafe { arr.push_raw(dependency) };
        assert_eq!(arr.as_slice().len(), 1);

        let arr_copy = arr.clone();
        assert_eq!(arr_copy.as_slice().len(), 1);
        assert!(std::ptr::eq(
            arr_copy.as_slice().last().unwrap().name,
            arr.as_slice().last().unwrap().name
        ));
    }
}
