/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::apk_name::ApkName;
use crate::{generate_apk_array_drop_impl, generate_apk_array_impl};
use apk_tools_sys::{apk_name, apk_name_array};

#[derive(Debug)]
pub struct ApkNameArray {
    inner: &'static mut apk_name_array,
    is_owned: bool,
}

generate_apk_array_impl!(ApkNameArray, apk_name_array, *mut apk_name);
generate_apk_array_drop_impl!(ApkNameArray);

impl ApkNameArray {
    pub fn iter(&self) -> impl Iterator<Item = ApkName> + use<'_> {
        self.as_slice()
            .iter()
            .filter_map(|&name| ApkName::from_ptr(name))
    }
}

#[cfg(test)]
mod test {
    use super::ApkNameArray;
    use crate::global_allocator::GlobalAllocator;

    #[test]
    fn test_arr_as_slice() {
        let raw = ApkNameArray::new_raw_array();
        let mut arr = ApkNameArray::from_ptr_owned(raw).unwrap();
        assert_eq!(arr.as_slice().len(), 0);
        let name = GlobalAllocator::get_apk_name("test");
        unsafe { arr.push_raw(name) };
        assert_eq!(arr.as_slice().len(), 1);
    }

    #[test]
    fn test_arr_copy() {
        let raw = ApkNameArray::new_raw_array();
        let mut arr = ApkNameArray::from_ptr_owned(raw).unwrap();
        let name = GlobalAllocator::get_apk_name("test");
        unsafe { arr.push_raw(name) };
        assert_eq!(arr.as_slice().len(), 1);

        let arr_copy = arr.clone();
        assert_eq!(arr_copy.as_slice().len(), 1);
        // assert!(std::ptr::eq(
        //     arr_copy.as_slice().last().unwrap(),
        //     arr.as_slice().last().unwrap()
        // ));
    }

    #[test]
    fn test_arr_iter() {
        let raw = ApkNameArray::new_raw_array();
        let mut arr = ApkNameArray::from_ptr_owned(raw).unwrap();
        let first_str = "hello";
        let second_str = "world";
        let first_name = GlobalAllocator::get_apk_name(first_str);
        let second_name = GlobalAllocator::get_apk_name(second_str);

        unsafe {
            arr.push_raw(first_name);
        }
        unsafe {
            arr.push_raw(second_name);
        }

        assert_eq!(arr.iter().next().unwrap().name(), first_str);
        assert_eq!(arr.iter().last().unwrap().name(), second_str);
    }
}
