/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::generate_apk_array_impl;
use apk_tools_sys::{_apk_array_truncate, apk_array, apk_string_array};
use std::ffi::{CStr, CString, c_char};

#[derive(Debug)]
pub struct ApkStringArray {
    inner: &'static mut apk_string_array,
    is_owned: bool,
}

generate_apk_array_impl!(ApkStringArray, apk_string_array, *mut c_char);

impl ApkStringArray {
    pub fn push(&mut self, value: &str) {
        let cstr = CString::new(value).unwrap();
        unsafe {
            self.push_raw(cstr.into_raw());
        }
    }

    pub fn iter(&self) -> impl Iterator<Item = &str> + use<'_> {
        self.as_slice()
            .iter()
            .map(|ptr| unsafe { CStr::from_ptr(*ptr).to_str().unwrap() })
    }

    pub fn steal(mut self) -> *mut apk_string_array {
        self.is_owned = false;
        self.inner
    }

    pub fn clear(&mut self) {
        if self.is_empty() {
            return;
        }

        for x in self.as_mut_slice() {
            unsafe { drop(CString::from_raw(*x)) };
        }

        let inner = self.inner as *mut apk_string_array;
        unsafe {
            _apk_array_truncate(inner as *mut apk_array, 0);
            self.inner = inner.as_mut().unwrap();
        }
    }
}

impl Drop for ApkStringArray {
    fn drop(&mut self) {
        if self.is_owned {
            self.clear();
            unsafe {
                Self::resize(&mut (self.inner as *mut _), 0, 0);
            }
        }
    }
}

#[cfg(test)]
mod test {
    use crate::ApkStringArray;

    #[test]
    fn test_push() {
        let mut arr = ApkStringArray::new();
        let expected = "test";
        arr.push(expected);
        assert!(!arr.is_empty());
        assert_eq!(arr.iter().next(), Some(expected));
    }
}
