/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

pub mod apk_change_array;
pub mod apk_dependency_array;
pub mod apk_name_array;
pub mod apk_provider_array;
pub mod apk_string_array;

pub use apk_change_array::*;
pub use apk_dependency_array::*;
pub use apk_name_array::*;
pub use apk_provider_array::*;
pub use apk_string_array::*;

#[macro_export]
macro_rules! generate_apk_array_impl {
    ($name:ident, $apk_array_type:ty, $array_element_type:ty) => {
        use apk_tools_sys::_apk_array_len;
        use paste::paste;

        impl $name {
            pub fn new() -> Self {
                Self {
                    inner: Self::new_raw_array(),
                    is_owned: true,
                }
            }

            #[allow(clippy::not_unsafe_ptr_arg_deref)]
            pub fn from_ptr(array: *mut $apk_array_type) -> Option<Self> {
                if array.is_null() {
                    return None;
                }

                Some(Self {
                    inner: unsafe { array.as_mut().unwrap() },
                    is_owned: false,
                })
            }

            pub fn from_ptr_owned(array: *mut $apk_array_type) -> Option<Self> {
                Self::from_ptr(array).map(|mut a| {
                    a.is_owned = true;
                    a
                })
            }

            pub fn len(&self) -> u32 {
                unsafe { _apk_array_len(&self.inner.hdr) }
            }

            pub fn as_slice(&self) -> &[$array_element_type] {
                unsafe { self.inner.item.as_slice(self.len() as usize) }
            }

            pub fn as_mut_slice(&mut self) -> &mut [$array_element_type] {
                unsafe { self.inner.item.as_mut_slice(self.len() as usize) }
            }

            pub fn is_empty(&self) -> bool {
                self.len() == 0
            }

            #[allow(dead_code)]
            pub(crate) fn as_mut(&mut self) -> &mut $apk_array_type {
                self.inner
            }

            fn new_raw_array() -> &'static mut $apk_array_type {
                let mut arr: *mut $apk_array_type = std::ptr::null_mut();
                unsafe {
                    paste! {
                        apk_tools_sys::[<$apk_array_type _init>](&mut arr);
                    }
                }
                unsafe { arr.as_mut().unwrap() }
            }

            #[allow(dead_code)]
            unsafe fn resize(a: *mut *mut $apk_array_type, size: usize, capacity: usize) {
                unsafe {
                    paste! {
                        apk_tools_sys::[<$apk_array_type _resize>](a, size, capacity);
                    }
                }
            }

            #[allow(dead_code)]
            unsafe fn add(
                a: *mut *mut $apk_array_type,
                item: $array_element_type,
            ) -> *mut $array_element_type {
                unsafe {
                    paste! {
                        apk_tools_sys::[<$apk_array_type _add>](a, item)
                    }
                }
            }

            #[allow(dead_code)]
            unsafe fn copy(a: *mut *mut $apk_array_type, b: *mut $apk_array_type) {
                unsafe {
                    paste! {
                        apk_tools_sys::[<$apk_array_type _copy>](a, b);
                    }
                }
            }

            #[allow(dead_code)]
            unsafe fn push_raw(&mut self, element: $array_element_type) {
                unsafe {
                    let mut inner = self.inner as *mut _;
                    Self::add(&mut inner, element);
                    self.inner = inner.as_mut().unwrap();
                }
            }
        }

        impl Clone for $name {
            fn clone(&self) -> Self {
                let mut other = Self::new_raw_array() as *mut _;

                unsafe {
                    Self::copy(&mut other, self.inner as *const _ as *mut _);
                }

                let mut ret = Self::from_ptr(other).unwrap();
                ret.is_owned = true;
                ret
            }
        }

        impl PartialEq for $name {
            fn eq(&self, other: &Self) -> bool {
                std::ptr::eq(self.inner, other.inner)
            }
        }

        impl Default for $name {
            fn default() -> Self {
                Self::new()
            }
        }

        #[cfg(test)]
        mod test_array {
            use super::$name;

            #[test]
            fn test_arr_from_ptr() {
                assert_eq!($name::from_ptr(std::ptr::null_mut()), None);

                let raw = $name::new_raw_array();
                let arr = $name::from_ptr_owned(raw);
                assert!(arr.is_some());
            }

            #[test]
            fn test_arr_resize() {
                let raw = $name::new_raw_array();
                let arr = $name::from_ptr(raw).unwrap();
                let mut inner = arr.inner as *mut _;
                let size = 0;
                unsafe {
                    $name::resize(&mut inner, size, size);
                }
                assert_eq!(arr.as_slice().len(), size);
            }
        }
    };
}

#[macro_export]
macro_rules! generate_apk_array_drop_impl {
    ($name:ident) => {
        impl Drop for $name {
            fn drop(&mut self) {
                if self.is_owned {
                    unsafe {
                        Self::resize(&mut (self.inner as *mut _), 0, 0);
                    }
                }
            }
        }
    };
}
