/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::apk_blob::ApkBlob;
use apk_tools_sys::{apk_blob_t, apk_name, apk_package, strlen};
use std::alloc::{Layout, alloc};
use std::ffi::CString;
use std::sync::{LazyLock, Mutex};

static ALLOCATOR: LazyLock<Mutex<GlobalAllocator>> =
    LazyLock::new(|| Mutex::new(GlobalAllocator::new()));

#[derive(Default)]
pub struct GlobalAllocator {
    names: Vec<*mut apk_name>,
    blobs: Vec<ApkBlob>,
    packages: Vec<apk_package>,
    c_strings: Vec<*mut core::ffi::c_char>,
}

unsafe impl Send for GlobalAllocator {}
unsafe impl Sync for GlobalAllocator {}

impl GlobalAllocator {
    fn new() -> Self {
        let mut ret = Self::default();

        ret.names.reserve(1000);
        ret.blobs.reserve(1000);
        ret.packages.reserve(1000);
        ret.c_strings.reserve(1000);

        ret
    }

    pub fn get_apk_name(name: &str) -> *mut apk_name {
        let mut instance = ALLOCATOR.lock().unwrap();
        let ptr = CString::new(name).unwrap().into_raw();
        let strlen = unsafe { strlen(ptr) as usize };

        let layout =
            Layout::from_size_align(size_of::<apk_name>() + strlen + 1, align_of::<apk_name>())
                .unwrap();
        let value = unsafe { alloc(layout) } as *mut apk_name;
        let name_ptr = unsafe { (*value).name.as_mut_ptr() };

        unsafe { std::ptr::copy_nonoverlapping(ptr, name_ptr, strlen + 1) };
        unsafe {
            let _ = CString::from_raw(ptr);
        };

        instance.names.push(value);
        value
    }

    pub fn get_apk_blob(content: &str) -> *mut apk_blob_t {
        let mut instance = ALLOCATOR.lock().unwrap();

        let blob = ApkBlob::new(content);
        instance.blobs.push(blob);
        instance.blobs.last_mut().unwrap().as_mut()
    }

    pub fn get_apk_package() -> *mut apk_package {
        let mut instance = ALLOCATOR.lock().unwrap();

        let package = apk_package::default();
        instance.packages.push(package);
        instance.packages.last_mut().unwrap() as *mut _
    }

    // pub fn get_c_string(string: &str) -> *mut core::ffi::c_char {
    //     let mut instance = ALLOCATOR.lock().unwrap();
    //
    //     let c_string = CString::new(string).unwrap();
    //     let raw = c_string.into_raw();
    //     instance.c_strings.push(raw);
    //     raw
    // }
}

impl Drop for GlobalAllocator {
    fn drop(&mut self) {
        for name in self.names.drain(..) {
            unsafe { drop(CString::from_raw((*name).name.as_mut_ptr())) }
        }

        for str in self.c_strings.drain(..) {
            unsafe { drop(CString::from_raw(str)) }
        }
    }
}
