/*
   Copyright: 2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use crate::apk_result::ApkResult;
use anyhow::{Result, anyhow, bail};
use apk_tools_sys::{
    apk_blob_pull_dep, apk_blob_t, apk_cache_download, apk_change_array_init, apk_changeset,
    apk_ctx, apk_database, apk_db_check_world, apk_db_close, apk_db_get_file_owner, apk_db_init,
    apk_db_open, apk_db_query_name, apk_db_select_repo, apk_dependency, apk_dependency_array,
    apk_hash_get_hashed, apk_hash_item, apk_name, apk_package, apk_repository, apk_solver_commit,
    apk_solver_commit_changeset, apk_solver_solve,
};
use std::any::Any;
use std::fmt::Debug;
use std::os::raw::c_ushort;
use std::ptr::null_mut;

pub(crate) trait ApkDatabaseProvider: Any + Debug {
    fn apk_blob_pull_dep(&mut self, name: &mut apk_blob_t) -> apk_dependency;
    fn apk_cache_download_pkg(
        &mut self,
        repository: *mut apk_repository,
        package: *mut apk_package,
    ) -> Result<()>;
    fn apk_cache_download_repo(&mut self, repo: &mut apk_repository) -> Result<()>;
    fn apk_db_check_world(&mut self, world: &mut apk_dependency_array) -> Result<()>;
    fn apk_db_get_file_owner(&mut self, filename: apk_blob_t) -> *mut apk_package;
    fn apk_db_init(&mut self, context: &mut apk_ctx);
    fn apk_db_open(&mut self) -> Result<()>;
    fn apk_db_query_name(&mut self, name: apk_blob_t) -> *mut apk_name;
    fn apk_db_select_repo(&mut self, pkg: *mut apk_package) -> *mut apk_repository;
    fn apk_hash_get_hashed(&mut self, key: apk_blob_t) -> apk_hash_item;
    fn apk_solver_commit(
        &mut self,
        solver_flags: u8,
        world: &mut apk_dependency_array,
    ) -> Result<()>;
    fn apk_solver_commit_changeset(
        &mut self,
        changeset: &mut apk_changeset,
        world: &mut apk_dependency_array,
    ) -> Result<()>;
    fn apk_solver_solve(
        &mut self,
        solver_flags: u8,
        world: &mut apk_dependency_array,
    ) -> Result<apk_changeset>;

    #[allow(clippy::mut_from_ref)]
    fn get_world(&self) -> &mut apk_dependency_array;
    fn get_repos(&mut self) -> Vec<&'static mut apk_repository>;
}

#[derive(Debug, Default)]
pub(crate) struct ApkDatabaseProviderImpl {
    inner: Box<apk_database>,
}

impl ApkDatabaseProviderImpl {
    pub fn new() -> Self {
        Self::default()
    }
}

impl ApkDatabaseProvider for ApkDatabaseProviderImpl {
    fn apk_blob_pull_dep(&mut self, name: &mut apk_blob_t) -> apk_dependency {
        let mut dependency = apk_dependency::default();
        unsafe {
            apk_blob_pull_dep(name, self.inner.as_mut(), &mut dependency, false);
        }
        dependency
    }

    fn apk_cache_download_pkg(
        &mut self,
        repository: *mut apk_repository,
        package: *mut apk_package,
    ) -> Result<()> {
        let res =
            unsafe { apk_cache_download(self.inner.as_mut(), repository, package, null_mut()) };

        if res == -libc::EALREADY {
            Ok(())
        } else {
            res.into_result()
        }
    }

    fn apk_cache_download_repo(&mut self, repo: &mut apk_repository) -> Result<()> {
        const DOWNLOAD_SUCCESS: i32 = 0;

        let result =
            unsafe { apk_cache_download(self.inner.as_mut(), repo, null_mut(), null_mut()) };

        if result == DOWNLOAD_SUCCESS {
            self.inner.repositories.updated += 1;
        } else if result != -libc::EALREADY {
            self.inner.repositories.unavailable += 1;
            return result.into_result();
        }

        Ok(())
    }

    fn apk_db_check_world(&mut self, world: &mut apk_dependency_array) -> Result<()> {
        unsafe { apk_db_check_world(self.inner.as_mut(), world) }
            .into_result()
            .map_err(|_| anyhow!("Missing repository tags; can't continue the upgrade!"))
    }

    fn apk_db_get_file_owner(&mut self, filename: apk_blob_t) -> *mut apk_package {
        unsafe { apk_db_get_file_owner(self.inner.as_mut(), filename) }
    }

    fn apk_db_init(&mut self, context: &mut apk_ctx) {
        unsafe { apk_db_init(self.inner.as_mut(), context) }
    }

    fn apk_db_open(&mut self) -> Result<()> {
        unsafe { apk_db_open(self.inner.as_mut()) }.into_result()
    }

    fn apk_db_query_name(&mut self, name: apk_blob_t) -> *mut apk_name {
        unsafe { apk_db_query_name(self.inner.as_mut(), name) }
    }

    fn apk_db_select_repo(&mut self, pkg: *mut apk_package) -> *mut apk_repository {
        unsafe { apk_db_select_repo(self.inner.as_mut(), pkg) }
    }

    fn apk_hash_get_hashed(&mut self, key: apk_blob_t) -> apk_hash_item {
        let available_names = &mut self.inner.available.names;
        let hash_key_fn = unsafe { available_names.ops.as_ref() }.unwrap().hash_key;
        let hash = unsafe { hash_key_fn.unwrap()(key) };
        unsafe { apk_hash_get_hashed(available_names, key, hash) }
    }

    fn apk_solver_commit(
        &mut self,
        solver_flags: u8,
        world: &mut apk_dependency_array,
    ) -> Result<()> {
        let num_errors =
            unsafe { apk_solver_commit(self.inner.as_mut(), solver_flags as c_ushort, world) };

        if num_errors == 0 {
            Ok(())
        } else {
            Err(anyhow!("{num_errors} errors occurred in changeset"))
        }
    }

    fn apk_solver_commit_changeset(
        &mut self,
        changeset: &mut apk_changeset,
        world: &mut apk_dependency_array,
    ) -> Result<()> {
        let num_errors =
            unsafe { apk_solver_commit_changeset(self.inner.as_mut(), changeset, world) };

        if num_errors == 0 {
            Ok(())
        } else {
            Err(anyhow!("{num_errors} errors occurred in changeset"))
        }
    }

    fn apk_solver_solve(
        &mut self,
        solver_flags: u8,
        world: &mut apk_dependency_array,
    ) -> Result<apk_changeset> {
        let mut changeset = apk_changeset::default();
        let errors = unsafe {
            apk_change_array_init(&mut changeset.changes);
            apk_solver_solve(
                self.inner.as_mut(),
                solver_flags as c_ushort,
                world,
                &mut changeset,
            )
        };

        if errors == 0 {
            Ok(changeset)
        } else {
            bail!("{errors} errors occurred while solving changeset");
        }
    }

    fn get_world(&self) -> &mut apk_dependency_array {
        unsafe { self.inner.world.as_mut() }.unwrap()
    }

    fn get_repos(&mut self) -> Vec<&'static mut apk_repository> {
        let num_repos = usize::try_from(self.inner.num_repos).unwrap();
        let mut ret = Vec::with_capacity(num_repos);

        for i in 0..num_repos {
            if let Some(repo) = self.inner.repos.get_mut(i) {
                let repo_raw = repo as *mut apk_repository;
                // FIXME: This technically *is* the incorrect lifetime, but we don't want rustc
                // to complain us holding a mutable ref to a repository while also calling
                // a apk_database method which requires mutable access to the provider
                let repo_static = unsafe { repo_raw.as_mut().unwrap() };
                ret.push(repo_static);
            }
        }

        ret
    }
}

impl Drop for ApkDatabaseProviderImpl {
    fn drop(&mut self) {
        if self.inner.open_complete() == 1 {
            unsafe { apk_db_close(self.inner.as_mut()) };
        }
    }
}
