/*
   Copyright: 2020-2025 Rasmus Thomsen <oss@cogitri.dev>
   SPDX-License-Identifier: GPL-3.0-or-later
*/

use std::env;
use std::path::{Path, PathBuf};

fn generate_bindings(include_paths: &Vec<&PathBuf>) {
    let out = std::env::var("OUT_DIR").unwrap();
    let out_dir = Path::new(&out);

    let mut builder = bindgen::Builder::default()
        .header("src/wrapper.h")
        .derive_copy(true)
        .derive_default(true)
        .size_t_is_usize(true)
        .formatter(bindgen::Formatter::Rustfmt)
        .impl_debug(true)
        .blocklist_type("max_align_t")
        .blocklist_type("apk_db_acl")
        .wrap_static_fns(true)
        .wrap_static_fns_path(out_dir.join("wrap_static_fns"))
        .parse_callbacks(Box::new(bindgen::CargoCallbacks::new()));

    for path in include_paths {
        builder = builder.clang_arg(format!("-I{}", path.to_str().unwrap()));
    }
    builder = builder.clang_arg("-DCRYPTO_USE_OPENSSL");

    let bindings = builder.generate().expect("Unable to generate bindings");

    bindings
        .write_to_file(out_dir.join("bindings.rs"))
        .expect("Couldn't write bindings!");

    let mut c_paths = include_paths.clone();
    let path = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
    c_paths.push(&path);

    cc::Build::new()
        .file(out_dir.join("wrap_static_fns.c"))
        .includes(c_paths)
        .define("CRYPTO_USE_OPENSSL", "1")
        .compile("wrap_static_fns");

    println!("cargo:rerun-if-changed=src/wrapper.h");
}

fn main() {
    let config = system_deps::Config::new()
        .probe()
        .expect("Couldn't find required system libraries");

    let mut include_paths = config.all_include_paths();
    let mut support_path = Path::new(env!("CARGO_MANIFEST_DIR")).to_path_buf();
    support_path.push("src");
    include_paths.push(&support_path);

    generate_bindings(&include_paths);
}
